import pandas as pd
import requests
from random import shuffle

# get json from wikidata raw url
# for example http://www.wikidata.org/entity/Q42493 turns into
# https://www.wikidata.org/wiki/Special:EntityData/Q42493.json
# then we get it via requests
def get_wikidata(original_url):
    url = original_url.replace("www.wikidata.org/wiki/", "www.wikidata.org/wiki/Special:EntityData/") + ".json"
    r = requests.get(url)
    rjson = r.json()
    try:
        return {list(rjson["entities"].values())[0]["labels"]["en"]["value"]}
    except:
        return pd.NA


df1 = pd.read_csv("gold_affiliation_annotations_2021.csv")
df2 = pd.read_csv("gold_affiliation_annotations_2022.csv")

# valid only
df2 = df2[df2.valid_affiliation_string == "yes"].drop(
    ["valid_affiliation_string", "openalex_institution_ids", "annotator"], axis=1
)

# sometimes incorrect, but there isn't a ror link and there is no wikidata link
incorrect_and_no_label = (df2.correct == "no") & df2.ror_labels_or_wikidata.apply(
    lambda s: pd.isna(s) or ("ror.org" not in s and "wikidata.org" not in s)
)
df2 = df2[~incorrect_and_no_label]

# custom processing for where there ror link
def process_ror_str(s):
    if not pd.isna(s):
        if "ror" in s:
            if "," in s:
                return set([i.strip() for i in s.split(",")])
            else:
                return set([s.strip()])
    return s


df2.loc[:, "ror_labels_or_wikidata"] = df2.ror_labels_or_wikidata.apply(process_ror_str)

# if ror_ids exists and it's correct and there is no ror_labels_or_wikidata -> move over ror_ids
correct_no_label = (df2.correct == "yes") & pd.isna(df2.ror_labels_or_wikidata) & ~pd.isna(df2.ror_ids)
df2.loc[correct_no_label, "ror_labels_or_wikidata"] = df2.loc[correct_no_label, "ror_ids"].apply(lambda s: set(eval(s)))

# get wikidata labels
has_wikidata = df2.ror_labels_or_wikidata.apply(lambda s: not pd.isna(s) and "wikidata.org" in s)
df2.loc[has_wikidata, "ror_labels_or_wikidata"] = df2.loc[has_wikidata, "ror_labels_or_wikidata"].apply(get_wikidata)

# for those still missing, just move over the raw_affiliation_string
still_missing = df2.ror_labels_or_wikidata.isna()
df2.loc[still_missing, "ror_labels_or_wikidata"] = df2.loc[still_missing, "raw_affiliation_string"]

# which ones aren't sets? some are HTML and some are strings
# drop HTML and convert strings to a set
def process_not_set(s):
    if "http" in s:
        return pd.NA
    else:
        return set([s])


not_set = df2.ror_labels_or_wikidata.apply(lambda s: not isinstance(s, set))
df2.loc[not_set, "ror_labels_or_wikidata"] = df2.loc[not_set, "ror_labels_or_wikidata"].apply(process_not_set)
df2.dropna(subset=["ror_labels_or_wikidata"], inplace=True)

# some of the sets have both ror and not ror
# if so, we should drop the rors only
def find_multi_source(s):
    has_ror = any(["ror.org" in i for i in s])
    if has_ror:
        return set([i for i in s if "ror.org" in i])
    else:
        return s


df2.loc[:, "ror_labels_or_wikidata"] = df2.ror_labels_or_wikidata.apply(find_multi_source)

# combine and save
df1_sub = df1[["original_affiliation", "mag_correct", "labels", "split"]]
df2_sub = df2[["raw_affiliation_string", "correct", "ror_labels_or_wikidata"]]
# randomly assign df2_sub.split one of:  'train' 'val' 'test'
n = df2_sub.shape[0]
train_val_test = ["train"] * int(n * 0.6) + ["val"] * int(n * 0.2)
train_val_test += ["test"] * (n - len(train_val_test))

shuffle(train_val_test)
df2_sub.loc[:, "split"] = train_val_test

df1_sub.columns = ["original_affiliation", "original_labels_correct", "labels", "split"]
df2_sub.columns = ["original_affiliation", "original_labels_correct", "labels", "split"]

# how correct are df1 and df2?
mag_correct = (df1_sub.original_labels_correct == "yes").mean()
openalex_correct = (df2_sub.original_labels_correct == "yes").mean()
print("MAG correct:", mag_correct)  # 0.77
print("OpenAlex correct:", openalex_correct)  # 0.83

df = pd.concat([df1_sub, df2_sub])

df.to_csv("gold_affiliation_annotations.csv", index=False)


rorless = df2_sub.labels.apply(lambda s: any([i for i in s if "ror.org" not in i]))



import requests
import zipfile
import os

# Make a GET request to the Zenodo API endpoint
response = requests.get("https://zenodo.org/api/records/?communities=ror-data&sort=mostrecent")

# Get the download URL of the most recent ROR record
download_url = response.json()["hits"]["hits"][0]["files"][0]["links"]["self"]
file_name = response.json()["hits"]["hits"][0]["files"][0]["key"]

# Download the record
response = requests.get(download_url)

# Save the record to a file
with open(file_name, "wb") as f:
    f.write(response.content)

# unzip the file_name and delete the zip file
with zipfile.ZipFile(file_name, "r") as zip_ref:
    zip_ref.extractall()
os.remove(file_name)

# get a list of all of the json files in the directory that that have the text 'ror-data' in them
# and sort it in order of most recent version to least recent version
json_files = [f for f in os.listdir(".") if os.path.isfile(f) and f.endswith(".json") and "ror-data" in f]
json_files.sort(reverse=True)

print("Downloaded this file: " + json_files[0])
print("Make sure to do these things and open a PR:")
print("(1) Update s2aff/conts.py with the new ROR_VERSION")
print(
    f"(2) Upload to the S3 bucket with the new ROR json file: aws s3 cp {json_files[0]} s3://ai2-s2-research-public/s2aff-release/"
)
print(
    f"(3) Delete from the S3 bucket with the old ROR json file: aws s3 rm s3://ai2-s2-research-public/s2aff-release/{json_files[1]}"
)



import json
import re
from collections import Counter, defaultdict
from operator import itemgetter
from random import shuffle

import numpy as np
import pandas as pd
from random import shuffle
from sklearn.feature_extraction.text import TfidfVectorizer
from collections import defaultdict, Counter
from s2aff.file_cache import cached_path
from operator import itemgetter
from s2aff.text import STOPWORDS, fix_text, INVERTED_ABBREVIATION_DICTIONARY
from s2aff.model import parse_ner_prediction
from s2aff.consts import (
    INSERT_EARLY_CANDIDATES_IND,
    MAX_INTERSECTION_DENOMINATOR,
    NS,
    PATHS,
    USE_PROB_WEIGHTS,
    WORD_MULTIPLIER,
    REINSERT_CUTOFF_FRAC,
    SCORE_BASED_EARLY_CUTOFF,
)
from s2aff.file_cache import cached_path


grid_extractor = re.compile(r"(grid\.\d{4,6}\.[0-9a-f]{1,2})")
isni_extractor = re.compile(r"(?=(\d{4}\s{0,1}\d{4}\s{0,1}\d{4}\s{0,1}[xX\d]{4}))")


def get_special_tokens_dict():
    special_tokens_dict = {
        "bos_token": "<s>",
        "eos_token": "</s>",
        "unk_token": "<unk>",
        "sep_token": "</s>",
        "pad_token": "<pad>",
        "cls_token": "<s>",
        "mask_token": "<mask>",
    }
    special_tokens_dict.update({f"additional_special_tokens": [f"<reserved-{i}>" for i in range(10)]})
    special_tokens_dict.update(
        {
            "[NAME]": special_tokens_dict["additional_special_tokens"][0],
            "[ACRONYM]": special_tokens_dict["additional_special_tokens"][1],
            "[ADDRESS]": special_tokens_dict["additional_special_tokens"][2],
            "[WIKI]": special_tokens_dict["additional_special_tokens"][3],
        }
    )
    return special_tokens_dict


def index_min(l, j_set):
    found_indices = []
    l_set = set(l)
    for j in j_set:
        if j in l_set:
            found_indices.append(l.index(j))
    if len(found_indices) > 0:
        return min(found_indices)
    else:
        return 100000  # approximate # of ROR elements


def sum_list_of_list_of_tuples(dicts, use_log=False):
    ret = defaultdict(float)
    for d in dicts:
        for k, v in d:
            if use_log:
                ret[k] += np.log1p(v)
            else:
                ret[k] += v
    return dict(ret)


class RORIndex:
    """An index for ROR (Research Organization Registry) entities.
    This can be used to find the most likely ROR entity for a given affiliation,
    as well as parse the predictions of the NER model.

    See https://ror.org/ for more details.

    Args:
        ror_data_path (str, optional): Location of the ROR data dump.
            Defaults to PATHS["ror_data"].
        country_info_path (str, optional): Location of additional data about countries
            that ROR does not have. Defaults to PATHS["country_info"].
        use_prob_weights (str, optional): Whether to use inverse word probability weights
            when using the Jaccard retrieval. Will effectively down-weight common affiliation words.
            Defaults to USE_PROB_WEIGHTS.
        word_multiplier (str, optional): The weight in
            word_weight * word_jaccard_score + ngram_jaccard_score. Defaults to WORD_MULTIPLIER.
        max_intersection_denominator (str, optional): Whether to modify how the denominator is
            calculated for Jaccard similarity. If true, only restrics the denominator to
            the set of words/ngrams found across all candidates. Defaults to MAX_INTERSECTION_DENOMINATOR.
        ns (set, optional): The set of ngram sizes to use when generating ngrams. Can be any combo of 3, 4, 5
            Defaults to {3}
    """

    def __init__(
        self,
        ror_data_path=PATHS["ror_data"],
        ror_edits_path=PATHS["ror_edits"],
        country_info_path=PATHS["country_info"],
        works_counts_path=PATHS["openalex_works_counts"],
        use_prob_weights=USE_PROB_WEIGHTS,
        word_multiplier=WORD_MULTIPLIER,
        max_intersection_denominator=MAX_INTERSECTION_DENOMINATOR,
        ns=NS,
        insert_early_candidates_ind=INSERT_EARLY_CANDIDATES_IND,
        reinsert_cutoff_frac=REINSERT_CUTOFF_FRAC,
        score_based_early_cutoff=SCORE_BASED_EARLY_CUTOFF,
    ):

        self.ror_data_path = cached_path(ror_data_path)
        self.country_info_path = cached_path(country_info_path)
        self.ror_edits_path = cached_path(ror_edits_path)
        works_counts = pd.read_csv(works_counts_path)
        self.works_counts = {i.ror: i.works_count for i in works_counts.itertuples()}
        self.use_prob_weights = use_prob_weights
        self.word_multiplier = word_multiplier
        self.max_intersection_denominator = max_intersection_denominator
        self.ns = ns
        self.insert_early_candidates_ind = insert_early_candidates_ind
        self.reinsert_cutoff_frac = reinsert_cutoff_frac
        self.score_based_early_cutoff = score_based_early_cutoff

        self.country_codes = pd.read_csv(self.country_info_path, sep="\t")
        self.country_codes_dict = {}
        for _, r in self.country_codes.iterrows():
            self.country_codes_dict[r["geonameid"]] = [r["Country"], r["ISO"], r["ISO3"], r["fips"]]

        with open(self.ror_data_path, "r") as f:
            ror = json.load(f)
            self.ror_dict = {i["id"]: i for i in ror if "name" in i}
            for key in self.ror_dict.keys():
                self.ror_dict[key]["works_count"] = self.works_counts.get(key, 0)

        self.grid_to_ror = {}
        self.isni_to_ror = {}
        for r in ror:
            if "GRID" in r["external_ids"]:
                self.grid_to_ror[r["external_ids"]["GRID"]["preferred"]] = r["id"]
            if 'ISNI' in r["external_ids"]:
                for isni in r["external_ids"]["ISNI"]["all"]:
                    self.isni_to_ror[isni.replace(' ', '')] = r["id"]

        # ROR database has some issues so we'll edit it directly
        with open(self.ror_edits_path, "r") as f:
            for line in f:
                line_json = json.loads(line)
                if line_json["ror_id"] in self.ror_dict:
                    print("Editing ROR database", line_json)
                    if line_json["action"] == "append":
                        if line_json["value"] not in self.ror_dict[line_json["ror_id"]][line_json["key"]]:
                            self.ror_dict[line_json["ror_id"]][line_json["key"]].append(line_json["value"])
                    elif line_json["action"] == "remove":
                        if line_json["value"] in self.ror_dict[line_json["ror_id"]][line_json["key"]]:
                            self.ror_dict[line_json["ror_id"]][line_json["key"]].remove(line_json["value"])

        # works_count came from OpenAlex and it has some errors
        # fixing them here by swapping works_count between parent ROR and child ROR
        # when the child ROR has suspiciously way more works_count AND the parent and child names are
        # name (location_1) and name (location_2)
        swaps_record = []
        for i in range(2):  # have to do this twice because there are multi-level hierarchies
            for ror_id in self.ror_dict.keys():
                parent_works_count = self.ror_dict[ror_id]["works_count"]
                parent_name = self.ror_dict[ror_id]["name"]
                relationships = self.ror_dict[ror_id]["relationships"]
                relationships = sorted(relationships, key=lambda x: -self.ror_dict[x["id"]]["works_count"])
                for rel in relationships:
                    if rel["type"] == "Child":
                        child_id = rel["id"]
                        child_entry = self.ror_dict[child_id]
                        child_name = child_entry["name"]
                        child_works_count = child_entry["works_count"]

                        # we only care about the ones where the names are the same EXCEPT anything in parentheses
                        name_no_parens = parent_name.split("(")[0].strip()
                        child_name_no_parens = child_name.split("(")[0].strip()
                        if name_no_parens == child_name_no_parens:
                            if child_works_count > parent_works_count:
                                self.ror_dict[ror_id]["works_count"] = child_works_count
                                self.ror_dict[child_id]["works_count"] = parent_works_count
                                swaps_record.append((parent_name, child_name))
                                break  # we sorted so this swap is only necessary once

        # we need some indices for fetching first order candidates
        ror_ngrams_inverted_index = defaultdict(set)
        ror_ngrams_lengths_index = {}

        ror_words_inverted_index = defaultdict(set)
        ror_words_lengths_index = {}

        ror_address_inverted_index = defaultdict(set)
        ror_city_inverted_index = defaultdict(set)
        ror_address_counter = Counter()

        inverse_dict_fixed = {}
        ror_name_direct_lookup = defaultdict(set)

        texts = []

        for r in ror:
            if "name" not in r:
                continue

            ror_id = r["id"]
            official_name = fix_text(r["name"]).lower().replace(",", "")
            aliases = [fix_text(i).lower().replace(",", "") for i in r["aliases"]]
            labels = [
                fix_text(i["label"]).lower().replace(",", "") for i in r["labels"]
            ]  # e.g. Sorbonne University has a label of Sorbonne Universités
            acronyms = [fix_text(i).lower().replace(",", "") for i in r["acronyms"]]
            all_names = [official_name] + aliases + labels + acronyms

            for name in all_names:
                if len(name) > 0:
                    ror_name_direct_lookup[name].add(ror_id)

            types_of_names = (
                ["official_name"]
                + [f"alias_{i}" for i in range(len(aliases))]
                + [f"label_{i}" for i in range(len(labels))]
                + [f"acronym_{i}" for i in range(len(acronyms))]
            )

            # this is the text we'll use later for the IDF weights
            text = " ".join(all_names)
            texts.append(text)

            for name, type_of_name in zip(all_names, types_of_names):
                ror_id_with_type = ror_id + "__" + type_of_name

                # get rid of stopwords
                name = " ".join([i for i in name.split(" ") if i not in STOPWORDS])
                inverse_dict_fixed[ror_id_with_type] = name

            # address
            if "addresses" in r and len(r["addresses"]) > 0:
                city = fix_text(r["addresses"][0]["city"]).lower().replace(",", "").split() or []
                state = fix_text(r["addresses"][0]["state"]).lower().replace(",", "").split() or []
                if r["addresses"][0]["state_code"] is not None:
                    state_code = [
                        fix_text(i).lower().replace(",", "") for i in r["addresses"][0]["state_code"].split("-")
                    ]
                else:
                    state_code = []

                # sometimes the state code is just a number - remove it if so
                state_code = [i for i in state_code if i.isalpha()]

                # additional country names and codes
                country_and_codes = self.country_codes_dict[r["addresses"][0]["country_geonames_id"]]
                if country_and_codes[0] == "China":
                    extras = ["pr", "prc"]
                else:
                    extras = []

                country_and_codes = [fix_text(i).lower().replace(",", "") for i in country_and_codes]

                # fix them up
                fixed_address_elements = set(city + state + country_and_codes + extras + state_code)
                fixed_address_elements = [i for i in fixed_address_elements if len(i) > 1 and i not in STOPWORDS]

                ror_address_counter.update([i for i in fixed_address_elements if len(i) > 3])

                for i in fixed_address_elements:
                    ror_address_inverted_index[i].add(ror_id)

                for i in city:
                    ror_city_inverted_index[i].add(ror_id)

        if self.use_prob_weights:
            vectorizer = TfidfVectorizer(
                min_df=1, analyzer="word", tokenizer=None, preprocessor=None, lowercase=False
            ).fit(texts)
            self.idf_lookup = {
                term: vectorizer.idf_[ind] for i, (term, ind) in enumerate(vectorizer.vocabulary_.items())
            }
            self.idf_lookup_min = min(self.idf_lookup.values())

            self.idf_weight = lambda i: self.idf_lookup.get(i, self.idf_lookup_min)
        else:
            self.idf_weight = None

        # make the indices
        for ror_id_with_type, name in inverse_dict_fixed.items():
            # ngrams
            name_ngrams, name_ngrams_weights = get_text_ngrams(name, weights_lookup_f=self.idf_weight, ns=self.ns)

            # one issue is when the same ngram appears in a long affiliation string
            # many times, it ends up being overweighted
            # hack mitigation: we keep each ngram only once, but the weight
            # is its max weight
            candidate_ngrams_unique = set()
            ngram_largest_weight = {}
            for ngram, weight in zip(name_ngrams, name_ngrams_weights):
                if ngram not in candidate_ngrams_unique:
                    candidate_ngrams_unique.add(ngram)
                    ngram_largest_weight[ngram] = np.maximum(weight, ngram_largest_weight.get(ngram, 0))

            if self.use_prob_weights:
                ror_ngrams_lengths_index[ror_id_with_type] = np.sum(
                    [ngram_largest_weight[i] for i in candidate_ngrams_unique]
                )
            else:
                ror_ngrams_lengths_index[ror_id_with_type] = len(candidate_ngrams_unique)

            for i in name_ngrams:
                ror_ngrams_inverted_index[i].add(ror_id_with_type)

            # words
            name_unigrams = set(name.split(" "))
            if self.use_prob_weights:
                ror_words_lengths_index[ror_id_with_type] = np.sum([self.idf_weight(i) for i in name_unigrams])
            else:
                ror_words_lengths_index[ror_id_with_type] = len(name_unigrams)
            for i in name_unigrams:
                ror_words_inverted_index[i].add(ror_id_with_type)

        self.word_index = {"inverted_index": dict(ror_words_inverted_index), "lengths_index": ror_words_lengths_index}
        self.ngram_index = {
            "inverted_index": dict(ror_ngrams_inverted_index),
            "lengths_index": ror_ngrams_lengths_index,
        }
        self.address_index = {
            "full_index": dict(ror_address_inverted_index),
            "city_index": dict(ror_city_inverted_index),
        }
        self.ror_address_counter = ror_address_counter
        self.ror_name_direct_lookup = ror_name_direct_lookup
        self.inverse_dict_fixed = inverse_dict_fixed

    def get_candidates_from_raw_affiliation(self, raw_affiliation, ner_predictor, look_for_grid_and_isni=True):
        """A wrapper function that puts the raw affiliation string through the
        NER predictor, parses the predicted output, and fetches the ROR candidates.

        Args:
            raw_affiliation (str): the raw affiliation string
            ner_predictor (NERPredictor): a model that can predict named affiliation entities
            look_for_grid_and_isni (bool): whether to look for GRID and ISNI IDs in the raw affiliation string
                Defalts to True. If found, just returns that candidate as the only one.

        Returns:
            candidates (list strings) - list of ROR inds in decreasing order of likelihood
            scores (list of flaots) - list of scores for each candidate

            Note that the candidates are sometimes *not* fully sorted by score because arbitrarily
            scored candidates are inserted heuristically into the list to increase recall.
        """
        if look_for_grid_and_isni:
            ror_from_grid = self.extract_grid_and_map_to_ror(raw_affiliation)
            ror_from_isni = self.extract_isni_and_map_to_ror(raw_affiliation)
            ror_from_grid_or_isni = ror_from_grid or ror_from_isni
            if ror_from_grid_or_isni is not None:
                return [ror_from_grid_or_isni], [1.0]
            
        ner_prediction = ner_predictor.predict([raw_affiliation])
        main, child, address, early_candidates = parse_ner_prediction(ner_prediction[0], self)
        candidates, scores = self.get_candidates_from_main_affiliation(main, address, early_candidates)
        return candidates, scores

    def get_candidates_from_main_affiliation(self, main, address="", early_candidates=[]):
        """Get ROR candidates via ngram and token Jaccard similarity.
        If you have a raw affiliation string, this is how to use this function:

        ```
        # instantiate NERPredictor and RORIndex here
        raw_affiliation = "Department of Philosophy, University of California, Berkeley, CA, USA"
        ner_prediction = ner_predictor.predict(raw_affiliation)
        main, child, address, early_candidates = parse_ner_prediction(ner_prediction, ror_index)
        candidates, scores = ror_index.get_candidates_from_main_affiliation(main, address, early_candidates)
        ```
        You can also use a convenience function as follows:
        ```
        candidates, scores = ror_index.get_candidates_from_raw_affiliation(raw_affiliation, ner_predictor)
        ```

        Args:
            main (str or list[str]): The string for the "main" part of the affiliation.
                For example, if the raw affiliation string is as above,
                then `main` should be "University of California".
                Note that this function expects that main and address have gone
                through `fix_text` but have not had stopwords removed nor are lowercased
            address (str, optional): The address part of the raw affiliation string.
                From above example, it would be "Berkeley, CA, USA". Defaults to "".
            early_candidates (list, optional): Some early candidate from parsing the
                raw affiliation string. Defaults to [].

        Returns:
            candidates (list strings) - list of ROR inds in decreasing order of likelihood
            scores (list of flaots) - list of scores for each candidate

            Note that the candidates are sometimes *not* fully sorted by score because arbitrarily
            scored candidates are inserted heuristically into the list to increase recall.
        """
        if type(main) is str:
            main = [main]
        ranked_before_dfs = []
        main_fixed_all = []
        for m in main:
            main_fixed = [i for i in m.lower().replace(",", "").split(" ") if i not in STOPWORDS]
            main_fixed = " ".join([INVERTED_ABBREVIATION_DICTIONARY.get(word, word) for word in main_fixed])
            main_fixed_all.append(main_fixed)
            if len(main_fixed) > 0:
                ranked_words_before = jaccard_word_nns(
                    main_fixed,
                    self.word_index,
                    idf_weight=self.idf_weight,
                    max_intersection_denominator=self.max_intersection_denominator,
                    word_multiplier=self.word_multiplier,
                    sort=False,
                )
                ranked_ngrams_before = jaccard_ngram_nns(
                    main_fixed,
                    self.ngram_index,
                    ns=self.ns,
                    idf_weight=self.idf_weight,
                    max_intersection_denominator=self.max_intersection_denominator,
                    sort=False,
                )
                ranked_before = sum_list_of_list_of_tuples(
                    [ranked_words_before, ranked_ngrams_before], use_log=self.word_multiplier == "log"
                )
                if len(ranked_before) > 0:
                    ranked_before_df = pd.DataFrame(ranked_before.items()).set_index(0)
                    if self.word_multiplier == "log":
                        ranked_before_df = ranked_before_df[
                            ranked_before_df[1] > np.log1p(self.score_based_early_cutoff)
                        ]
                    else:
                        ranked_before_df = ranked_before_df[ranked_before_df[1] > self.score_based_early_cutoff]
                    ranked_before_dfs.append(ranked_before_df)
        if len(ranked_before_dfs) > 0:
            x = pd.concat(ranked_before_dfs, axis=1).fillna(0).max(1)
        else:
            return [], []
        ranked_before = list(zip(x.index, x))
        ranked_before = sorted(ranked_before, key=itemgetter(1), reverse=True)

        # we have multiple appearances of the same grid id
        # keep only the first appearance in the ranked list
        ranked_unique = []
        seen_rors = set()
        for ror_id, score in ranked_before:
            ror_id = ror_id.split("__")[0]
            if ror_id not in seen_rors:
                ranked_unique.append((ror_id, score))
                seen_rors.add(ror_id)

        # insert early candidates in position self.insert_early_candidates_ind if not seen already
        if self.insert_early_candidates_ind is not None:
            early_candidates_tuples = [(i, -0.1) for i in early_candidates if i not in seen_rors]
            ranked_unique = (
                ranked_unique[: self.insert_early_candidates_ind]
                + early_candidates_tuples
                + ranked_unique[self.insert_early_candidates_ind :]
            )

        if len(ranked_unique) == 0:
            return [], []

        # if we have address text, we can find the RORs that have at least one word in that text
        if address != "" and len(address) > 0:
            if type(address) is list:
                address = " ".join(address)
            address_fixed_tokens = set([i for i in address.lower().replace(",", "").split(" ") if i not in STOPWORDS])

            inverted = self.address_index["full_index"]
            acceptable_rors = set()
            for unigram in address_fixed_tokens:
                if unigram in inverted:
                    acceptable_rors.update(inverted[unigram])

            ranked_after_address_filter = [i for i in ranked_unique if i[0] in acceptable_rors]

            if len(ranked_after_address_filter) == 0:
                ranked_after_address_filter = ranked_unique
            else:
                # reinsert top from ranked_unique that were removed, just in case the address was wrong
                ranked_removed = [i for i in ranked_unique if i[0] not in acceptable_rors]
                if len(ranked_removed) > 0:
                    # reinsert the ones that all are close to top score
                    ranked_removed_top_score = self.reinsert_cutoff_frac * ranked_after_address_filter[0][1]
                    # artificially setting the score to distinguish from "real" top
                    ranked_removed_to_reinsert = [
                        (i[0], -0.15) for i in ranked_removed if i[1] >= ranked_removed_top_score
                    ]
                    if self.insert_early_candidates_ind is not None:
                        insert_ind = len(early_candidates_tuples) + self.insert_early_candidates_ind + 10
                    else:
                        insert_ind = len(early_candidates_tuples) + 10
                    ranked_after_address_filter = (
                        ranked_after_address_filter[:insert_ind]
                        + ranked_removed_to_reinsert
                        + ranked_after_address_filter[insert_ind:]
                    )
        else:
            ranked_after_address_filter = ranked_unique

        candidates, scores = list(zip(*ranked_after_address_filter))

        return candidates, scores

    def parse_ror_entry_into_single_string(
        self, ror_id, use_separator_tokens=True, shuffle_names=False, special_tokens_dict=get_special_tokens_dict()
    ):
        if ror_id not in self.ror_dict:
            # assuming that this is just not a ror_id and instead a string of some affiliation
            return "[NAME] " + ror_id
        ror_entry = self.ror_dict[ror_id]
        return parse_ror_entry_into_single_string(
            ror_entry, self.country_codes_dict, special_tokens_dict, use_separator_tokens, shuffle_names, use_wiki=False
        )
        
    def extract_grid_and_map_to_ror(self, s):
        extracted_grids = grid_extractor.findall(s)
        if len(extracted_grids) == 0:
            return None
        else:
            extracted_grid_1 = extracted_grids[0]
            extracted_ror_1 = self.grid_to_ror.get(extracted_grid_1, None)
            if len(extracted_grid_1.split('.')[-1]) == 2:
                extracted_grid_2 = extracted_grid_1[:-1]
                extracted_ror_2 = self.grid_to_ror.get(extracted_grid_2, None)
                if extracted_ror_1 and extracted_ror_2:
                    return None
                else:
                    return extracted_ror_1 or extracted_ror_2
            else:
                return extracted_ror_1

    def extract_isni_and_map_to_ror(self, s):
        extracted_isnis = isni_extractor.findall(s)
        # look up all extracted ISNIs and return the ror if exactly one ROR is found
        found_rors = [self.isni_to_ror.get(isni.replace(' ', ''), None) for isni in extracted_isnis]
        filtered_rors = [x for x in found_rors if x is not None]
        if len(filtered_rors) == 1:
            return filtered_rors[0]
        else:
            return None



def parse_ror_entry_into_single_string(
    ror_entry,
    country_codes_dict,
    special_tokens_dict,
    use_separator_tokens=True,
    shuffle_names=False,
    use_wiki=True,
):

    official_name = fix_text(ror_entry["name"])
    aliases = [fix_text(i) for i in ror_entry["aliases"]]
    labels = [
        fix_text(i["label"]) for i in ror_entry["labels"]
    ]  # e.g. Sorbonne University has a label of Sorbonne Universités
    acronyms = [fix_text(i) for i in ror_entry["acronyms"]]
    all_names = [official_name] + aliases + labels + acronyms

    if use_separator_tokens:
        types_of_names = (
            [f"{special_tokens_dict.get('[NAME]')}"]
            + [f"{special_tokens_dict.get('[NAME]')}" for i in range(len(aliases))]
            + [f"{special_tokens_dict.get('[NAME]')}" for i in range(len(labels))]
            + [f"{special_tokens_dict.get('[ACRONYM]')}" for i in range(len(acronyms))]
        )

        names_and_types = list(zip(all_names, types_of_names))
        if shuffle_names:
            shuffle(names_and_types)
        output = " ".join(f"{b} {a}" for a, b in names_and_types)
    else:
        if shuffle_names:
            shuffle(all_names)
        output = " ; ".join(all_names)

    if "addresses" in ror_entry and len(ror_entry["addresses"]) > 0:
        city = ror_entry["addresses"][0]["city"] or ""
        state = ror_entry["addresses"][0]["state"] or ""
        if ror_entry["addresses"][0]["state_code"] is not None:
            state_code = ror_entry["addresses"][0]["state_code"].split("-")
        else:
            state_code = []
        state_code = [i for i in state_code if i.isalpha()]
        country_and_codes = country_codes_dict[ror_entry["addresses"][0]["country_geonames_id"]]
        if country_and_codes[0] == "China":
            extras = ["PR", "PRC"]
        else:
            extras = []
        fixed_address_elements = []
        seen = set()
        for i in [city, state] + state_code + country_and_codes + extras:
            i_fix = fix_text(i)
            if i_fix not in seen and len(i_fix) > 0:
                fixed_address_elements.append(i_fix)
                seen.add(i_fix)

        address = " ".join(fixed_address_elements)
        if use_separator_tokens:
            output += f" {special_tokens_dict.get('[ADDRESS]')} {address}"
        else:
            output += f" ; {address}"

    if "wikipedia_page" in ror_entry and use_wiki:
        output += f" {special_tokens_dict.get('[WIKI]')} {' '.join(ror_entry['wikipedia_page'])}"

    return output


def jaccard_ngram_nns(
    candidate,
    ngram_index,
    ns=NS,
    idf_weight=None,
    max_intersection_denominator=MAX_INTERSECTION_DENOMINATOR,
    sort=False,
):

    lengths = ngram_index["lengths_index"]
    inverted = ngram_index["inverted_index"]

    intersections = Counter()
    candidate_ngrams, candidate_ngram_weights = get_text_ngrams(candidate, weights_lookup_f=idf_weight, ns=ns)

    # one issue is when the same ngram appears in a long affiliation string
    # many times, it ends up being overweighted
    # hack mitigation: we keep each ngram only once, but the weight
    # is its max weight
    candidate_ngrams_unique = set()
    ngram_largest_weight = {}
    for ngram, weight in zip(candidate_ngrams, candidate_ngram_weights):
        if ngram not in candidate_ngrams_unique:
            candidate_ngrams_unique.add(ngram)
            ngram_largest_weight[ngram] = np.maximum(weight, ngram_largest_weight.get(ngram, 0))

    intersections = defaultdict(float)
    candidate_ngrams_in_inverted = []
    weights_in_inverted = []
    for ng in candidate_ngrams_unique:
        if ng in inverted:
            weight = ngram_largest_weight[ng]
            for ror_id in inverted[ng]:
                intersections[ror_id] += weight
            weights_in_inverted.append(weight)
            candidate_ngrams_in_inverted.append(ng)

    if idf_weight is not None:
        num_candidate_ngrams = np.sum(candidate_ngram_weights)
        num_relevant_candidate_ngrams = np.sum(weights_in_inverted)
    else:
        num_candidate_ngrams = len(candidate_ngrams)
        num_relevant_candidate_ngrams = len(candidate_ngrams_in_inverted)

    if max_intersection_denominator:
        jaccards = [
            (grid, intersection / (num_relevant_candidate_ngrams + lengths[grid] - intersection))
            for grid, intersection in intersections.items()
        ]
    else:
        jaccards = [
            (grid, intersection / (num_candidate_ngrams + lengths[grid] - intersection))
            for grid, intersection in intersections.items()
        ]

    if sort:
        jaccards = sorted(jaccards, key=itemgetter(1), reverse=True)

    return jaccards


def jaccard_word_nns(
    candidate,
    word_index,
    idf_weight=None,
    max_intersection_denominator=MAX_INTERSECTION_DENOMINATOR,
    word_multiplier=WORD_MULTIPLIER,
    sort=False,
):
    if word_multiplier == "log":
        word_multiplier = 1

    lengths = word_index["lengths_index"]
    inverted = word_index["inverted_index"]

    candidate_unigrams = set(candidate.split())

    use_prob_weights = idf_weight is not None

    if use_prob_weights:
        unigram_inv_probs = {i: idf_weight(i) for i in candidate_unigrams}

    intersections = defaultdict(float)
    all_matched_unigrams = set()
    for unigram in candidate_unigrams:
        if unigram in inverted:
            if use_prob_weights:
                weight = unigram_inv_probs[unigram]
            else:
                weight = 1
            for ror_id in inverted[unigram]:
                intersections[ror_id] += weight
            all_matched_unigrams.add(unigram)

    if use_prob_weights:
        num_candidate_unigrams = np.sum(list(unigram_inv_probs.values()))
        num_relevant_candidate_unigrams = np.sum([unigram_inv_probs[i] for i in all_matched_unigrams])
    else:
        num_candidate_unigrams = len(candidate_unigrams)
        num_relevant_candidate_unigrams = len(all_matched_unigrams)

    if max_intersection_denominator:
        jaccards = [
            (
                ror_id,
                word_multiplier * intersection / (num_relevant_candidate_unigrams + lengths[ror_id] - intersection),
            )
            for ror_id, intersection in intersections.items()
        ]
    else:
        jaccards = [
            (ror_id, word_multiplier * intersection / (num_candidate_unigrams + lengths[ror_id] - intersection))
            for ror_id, intersection in intersections.items()
        ]

    if sort:
        jaccards = sorted(jaccards, key=itemgetter(1), reverse=True)

    return jaccards


def get_text_ngrams(text, weights_lookup_f=None, ns={3, 4, 5}):
    """
    Get character bigrams, trigrams, quadgrams, and optionally unigrams for a piece of text.
    Note: respects word boundaries

    Parameters
    ----------
    text: string
        the text to get ngrams for
    use_unigrams: bool
        whether or not to include unigrams
    stopwords: Set
        The set of stopwords to filter out before computing character ngrams

    Returns
    -------
    Counter: the ngrams present in the text
    """
    if text is None or len(text) == 0:
        return [], []

    words = text.split(" ")

    if weights_lookup_f is not None:
        unigram_inv_probs = [weights_lookup_f(i) for i in words]
    else:
        unigram_inv_probs = [1] * len(words)

    text = " ".join(words)

    ngrams = []
    weights = []

    if 3 in ns:
        trigrams = map(
            lambda x: "".join(x),
            filter(lambda x: " " not in x, zip(text, text[1:], text[2:])),
        )
        trigram_weights = sum(
            [[j / max(len(i) - 2, 1e-6)] * (len(i) - 2) for i, j in zip(words, unigram_inv_probs)], []
        )
        ngrams += list(trigrams)
        weights += trigram_weights

    if 4 in ns:
        quadgrams = map(
            lambda x: "".join(x),
            filter(lambda x: " " not in x, zip(text, text[1:], text[2:], text[3:])),
        )
        quadgram_weights = sum(
            [[j / max(len(i) - 3, 1e-6)] * (len(i) - 3) for i, j in zip(words, unigram_inv_probs)], []
        )
        ngrams += list(quadgrams)
        weights += quadgram_weights

    if 5 in ns:
        fivegrams = map(
            lambda x: "".join(x),
            filter(lambda x: " " not in x, zip(text, text[1:], text[2:], text[3:], text[4:])),
        )
        fivegram_weights = sum(
            [[j / max(len(i) - 4, 1e-6)] * (len(i) - 4) for i, j in zip(words, unigram_inv_probs)], []
        )
        ngrams += list(fivegrams)
        weights += fivegram_weights

    return ngrams, weights



import logging
import os
import ntpath
from pathlib import Path

logger = logging.getLogger("s2aff")

try:
    PROJECT_ROOT_PATH = os.path.abspath(os.path.join(__file__, os.pardir, os.pardir))
except NameError:
    PROJECT_ROOT_PATH = os.path.abspath(os.path.join(os.getcwd()))


CACHE_ROOT = Path(os.getenv("S2AFF_CACHE", str(Path.home() / ".s2aff")))

ROR_VERSION = "v1.25-2023-05-11"

PATHS = {
    "ner_training_data": os.path.join(PROJECT_ROOT_PATH, "data", "ner_training_data.pickle"),
    "ror_data": os.path.join(PROJECT_ROOT_PATH, "data", f"{ROR_VERSION}-ror-data.json"),
    "country_info": os.path.join(PROJECT_ROOT_PATH, "data", "country_info.txt"),
    "openalex_works_counts": os.path.join(PROJECT_ROOT_PATH, "data", "openalex_works_counts.csv"),
    "gold_affiliation_annotations": os.path.join(PROJECT_ROOT_PATH, "data", "gold_affiliation_annotations.csv"),
    "ner_model": os.path.join(PROJECT_ROOT_PATH, "data", "ner_model"),
    "kenlm_model": os.path.join(PROJECT_ROOT_PATH, "data", "raw_affiliations_lowercased.binary"),
    "lightgbm_gold_features": os.path.join(PROJECT_ROOT_PATH, "data", "lightgbm_gold_features.pickle"),
    "lightgbm_gold_no_ror_features": os.path.join(PROJECT_ROOT_PATH, "data", "lightgbm_gold_no_ror_features.pickle"),
    "lightgbm_model": os.path.join(PROJECT_ROOT_PATH, "data", "lightgbm_model.booster"),
    "ror_edits": os.path.join(PROJECT_ROOT_PATH, "data", "ror_edits.jsonl"),
}

for key in PATHS.keys():
    if not os.path.exists(PATHS[key]):
        PATHS[key] = "https://s3-us-west-2.amazonaws.com/ai2-s2-research-public/s2aff-release/" + ntpath.basename(
            PATHS[key]
        )

USE_PROB_WEIGHTS = True
MAX_INTERSECTION_DENOMINATOR = True
WORD_MULTIPLIER = 0.3
NS = {3}  # character trigrams only
INSERT_EARLY_CANDIDATES_IND = 1  # insert early candidates at second place
REINSERT_CUTOFF_FRAC = 1.1  # reinsert candidates with score > cutoff
SCORE_BASED_EARLY_CUTOFF = 0.01  # anything with a score this low is removed before any kind of sorting is done



"""
Write integration tests for your model interface code here.

The TestCase class below is supplied a `container`
to each test method. This `container` object is a proxy to the
Dockerized application running your model. It exposes a single method:

```
predict_batch(instances: List[Instance]) -> List[Prediction]
```

To test your code, create `Instance`s and make normal `TestCase`
assertions against the returned `Prediction`s.

e.g.

```
def test_prediction(self, container):
    instances = [Instance(), Instance()]
    predictions = container.predict_batch(instances)

    self.assertEqual(len(instances), len(predictions)

    self.assertEqual(predictions[0].field1, "asdf")
    self.assertGreatEqual(predictions[1].field2, 2.0)
```
"""


import logging
import sys
import unittest

from .interface import Instance, Prediction


try:
    from timo_interface import with_timo_container
except ImportError as e:
    logging.warning(
        """
    This test can only be run by a TIMO test runner. No tests will run. 
    You may need to add this file to your project's pytest exclusions.
    """
    )
    sys.exit(0)


@with_timo_container
class TestInterfaceIntegration(unittest.TestCase):
    def test__predictions(self, container):
        instances = [
            Instance(raw_affiliation="The University of Texas at Austin"),
            Instance(raw_affiliation="Univ of Wash, Seattle, WA, 98115, U.S.A."),
        ]
        predictions = container.predict_batch(instances)

        self.assertEqual(predictions[0].ror_id, "https://ror.org/00hj54h04")
        self.assertEqual(predictions[1].ror_id, "https://ror.org/00cvxb145")



import os
import torch

# have to do this to avoid weird bugs
os.environ["TOKENIZERS_PARALLELISM"] = "false"
torch.multiprocessing.set_sharing_strategy("file_system")

import gc
import numpy as np
import lightgbm as lgb
import kenlm
from s2aff.text import fix_text
from s2aff.features import make_lightgbm_features, parse_ror_entry_into_single_string_lightgbm, FEATURE_NAMES
from simpletransformers.ner import NERModel, NERArgs
from blingfire import text_to_words
from s2aff.consts import PATHS
from s2aff.text import fix_text, CERTAINLY_MAIN


FEATURE_NAMES = list(FEATURE_NAMES)


def parse_ner_prediction(ner_prediction, ror_index):
    """Parse the NER prediction that comes out of a NERPredictor
    into main entity, child entities, address, and early candidates for the
    next retrieval stage.

    Args:
        ner_prediction (list of dicts): Output of a NERPredictor.predict. Looks like this:
            [{'Chinese': 'I-MAIN'},
                {'Academy': 'I-MAIN'},
                {'of': 'I-MAIN'},
                {'Sciences': 'I-MAIN'},
                {'PRC': 'I-ADDRESS'}]

                Note that it's one prediction and not a list of predictions.

    Returns:
        main (str): Main entity of the child affiliation string.
        child (list[str]): Child entities of the raw affiliation string.
        address (list): Address part of the raw affiliation string.
        early_candidates (list[str]): set of candidates based off child entities.
            These appear in the direct ROR look-up table.
    """
    main = ""
    child = ""
    main_child = ""
    address = ""
    previous_tag_suffix = "not a tag"
    for d in ner_prediction:
        word, tag = list(d.items())[0]
        # if a second affiliation starts, we'll just take the first one only
        # TODO: actually return the second affiliation?
        if "SPLIT" in tag and word == "and":  # sometimes the model decides other words are split words
            break
        elif "SPLIT" in tag:
            tag = "O"
        if tag != "O":
            tag_suffix = tag.split("-")[1]

            begin = tag.startswith("B") or (previous_tag_suffix != tag_suffix)

            if word != ",":
                word_to_append = word + " "
            else:
                word_to_append = ""
            if "MAIN" in tag or "CHILD" in tag:
                main_child += word_to_append
                if "MAIN" in tag:
                    main += ("BEGIN" if begin else "") + word_to_append
                else:
                    child += ("BEGIN" if begin else "") + word_to_append
            elif "ADDRESS" in tag:
                address += ("BEGIN" if begin else "") + word_to_append

            previous_tag_suffix = tag_suffix
        else:
            previous_tag_suffix = "O"  # this is how we break up multiple continuous children

    address = [i.strip() for i in address.split("BEGIN") if len(i) > 1]

    if len(main) > 0:
        main = [i.strip() for i in main.split("BEGIN") if len(i) > 1]
        child = [i.strip() for i in child.split("BEGIN") if len(i) > 1]
    elif len(main_child) > 0:
        main = [main_child.strip()]
        child = []
    elif len(address) > 0:
        main = address
        child = []
        address = []

    # check if any of the children have a "main" word and move to main
    child_new = []
    for i in child:
        child_split = set(i.split())
        if len(child_split.intersection(CERTAINLY_MAIN)) > 0:
            main.append(i)
        else:
            child_new.append(i)

    # we have direct look-up tables for both grid names and addresses
    # here we: check the children. if any appear in grid names lookup
    # we can add to early grid candidates
    early_candidates = []
    for c in child_new:
        c = c.lower()
        if c in ror_index.ror_name_direct_lookup:
            early_candidates.extend(list(ror_index.ror_name_direct_lookup[c]))

    # if something that is a main appears in grid addresses lookup
    # but not in the direct name look up
    # it is a sign that main was possibly incorrectly extracted.
    # To mitigate, we (a) remove these from the mains and
    # (b) add the children into mains to supplement
    main_in_address = False
    mains_to_remove = set()
    for m in main:
        if m.lower() in ror_index.ror_address_counter:
            main_in_address = True
            if m.lower() not in ror_index.ror_name_direct_lookup:
                mains_to_remove.add(m)

    if main_in_address:
        main.extend(child_new)

    # if a main is in the address lookup AND not in name lookup
    # we remove it and swap into address and children. it's probably just a mislabeled address
    for m in mains_to_remove:
        main.remove(m)
        if m not in child_new:
            child_new.append(m)
        if m not in address:
            address.append(m)

    # we might not have a main still. it just becomes child or address
    if len(main) == 0:
        if len(child_new) > 0:
            main = child_new
            child_new = []
        elif len(address) > 0:
            main = address
            address = []

    # join it all together
    return main, child_new, address, early_candidates


class NERPredictor:
    """Named Entity Recognition for affiliation strings.
    Uses SimpleTransformers under the hood.
    """

    def __init__(self, model_path=PATHS["ner_model"], model_type="roberta", use_cuda=True):
        self.model_path = model_path
        self.model_type = model_type
        self.use_cuda = use_cuda
        if self.model_path is not None:
            self.load_model(self.model_path, self.model_type)

    def load_model(self, model_path=PATHS["ner_model"], model_type="roberta"):
        """Load a model from disk.

        model_path (str, optional): Location of the saved NER model.
            Should be what is saved by SimpleTransformers NERModel. Defaults to PATHS["ner_model"].
        model_type (str, optional): Model type such as roberta or bert.
            If you don't know, check the config.json in the model directory. Defaults to "roberta".
        """
        self.model = NERModel(
            model_type,
            model_path,
            use_cuda=self.use_cuda,
            args={
                "use_multiprocessing": False,
                "use_multiprocessing_for_evaluation": False,
                "process_count": 1,
                "eval_batch_size": 8,
            },
        )

    def save_model(self, model_path=PATHS["ner_model"]):
        """Save model to disk

        Args:
            model_output_path (str, optional): Where to save the model.
                Defaults to PATHS["ner_model"].
        """
        self.model.save_model(output_dir=model_path, model=self.model)

    def delete_model(self):
        """Clears the model from GPU memory."""
        del self.model
        gc.collect()
        torch.cuda.empty_cache()

    def fit(
        self,
        df_train,
        df_validation=None,
        model_type="roberta",
        model_name="roberta-large",
        num_train_epochs=3,
        learning_rate=1e-5,
    ):
        """Fit the NER model.
        Uses SimpleTransformers under the hood.

        Args:
            df_train (pd.DataFrame): Training data. Assumes data has already been preprocessed
                with `fix_text` and tokenized with  `blingfire.text_to_words`, and then shaped
                into a dataframe with 3 columns: ['sentence_id', 'words', 'labels'].
            df_validation (pd.DataFrame): Same as above but for validation. None by default.
            model_type (str, optional): Model type such as roberta or bert.. Defaults to "roberta".
            model_name (str, optional): Specific model name of the model_class. Defaults to "roberta-large".
            num_train_epochs (int, optional): Number of training epochs. Defaults to 3.
            learning_rate (float, optional): Learning rate. Defaults to 1e-5.

        Returns:
            model: SimpleTransformers NERModel.
            result_vl: Validation metrics provided by SimpleTransformers.
        """
        # most of these are not part of the input params, but they could be if we wanted
        model_args = NERArgs()
        model_args.scheduler = "linear_schedule_with_warmup"
        model_args.num_train_epochs = num_train_epochs
        model_args.learning_rate = learning_rate
        model_args.hidden_dropout_prob = 0.3
        model_args.attention_probs_dropout_prob = 0.3
        model_args.evaluate_during_training = False
        model_args.reprocess_input_data = True
        model_args.overwrite_output_dir = True
        model_args.manual_seed = 4
        model_args.train_batch_size = 32
        model_args.eval_batch_size = 128
        model_args.use_multiprocessing = False
        model_args.use_multiprocessing_for_evaluation = False
        model_args.process_count = 1

        # train and save model
        custom_labels = list(set(df_train.labels))
        model = NERModel(model_type, model_name, labels=custom_labels, args=model_args, use_cuda=self.use_cuda)
        model.train_model(df_train)
        self.model = model.model

        if df_validation is not None:
            result_vl, _ = self.eval(df_validation)
        else:
            result_vl = None

        return result_vl

    def eval(self, df):
        result, model_outputs, _ = self.model.eval_model(df)
        return result, model_outputs

    def predict(self, texts):
        """Do NER on input affiliation string(s).

        Args:
            texts (str or list[str]): Affiliation string. Can be a list of strings, in which case
                the output will be a list of predictions.

        Returns:
            predictions: List of list of dicts. Looks like this if you pass in string.
                [{'Chinese': 'I-MAIN'},
                {'Academy': 'I-MAIN'},
                {'of': 'I-MAIN'},
                {'Sciences': 'I-MAIN'},
                {'PRC': 'I-ADDRESS'}]
            and will be a list of predictions if the input is a list of strings.
        """
        if type(texts) == str:
            texts = [texts]
            input_was_single_text = True
        else:
            input_was_single_text = False

        texts = [text_to_words(fix_text(text)) for text in texts]

        predictions = self.model.predict(texts)[0]  # [1] are the scores

        if input_was_single_text:
            return predictions[0]
        else:
            return predictions


class PairwiseRORLightGBMReranker:
    def __init__(
        self, ror_index, model_path=PATHS["lightgbm_model"], kenlm_model_path=PATHS["kenlm_model"], num_threads=0
    ):
        self.ror_index = ror_index
        self.model_path = model_path
        self.kenlm_model_path = kenlm_model_path
        self.load_model(model_path)
        self.lm = kenlm.LanguageModel(kenlm_model_path)
        self.num_threads = num_threads

        self.inds_to_check = [
            FEATURE_NAMES.index("names_frac_of_query_matched_in_text"),
            FEATURE_NAMES.index("acronyms_frac_of_query_matched_in_text"),
        ]
        self.city_ind = FEATURE_NAMES.index("city_frac_of_query_matched_in_text")
        self.num_threads = num_threads

    def load_model(self, model_path=PATHS["lightgbm_model"]):
        """Load a model from disk.

        model_path (str, optional): Location of the saved parwise classifier.
            Should be what is saved by SimpleTransformers ClassificationModel.
            Defaults to PATHS["reranker_model"].
        model_type (str, optional): Model type such as roberta or bert.
            If you don't know, check the config.json in the model directory. Defaults to "roberta".
        """
        self.model = lgb.Booster(model_file=model_path)

    def save_model(self, model_path=PATHS["lightgbm_model"]):
        """Save model to disk

        Args:
            model_output_path (str, optional): Where to save the model.
                Defaults to PATHS["lightgbm_model"].
        """
        self.model.save_model(output_dir=model_path, model=self.model)

    def delete_model(self):
        """Clears the model from GPU memory."""
        del self.model
        gc.collect()

    def predict(self, raw_affiliation, candidates, scores):
        """
        Given a list of candidates that are ROR ids, re-rank them using the trained model.

        Args:
            raw_affiliation (str): Raw affiliation string.
            candidates (list[str]): List of candidate ROR ids.

        Returns:
            reranked_candidates (np.array[str]): Array of candidate ROR ids
            reranked_scores (np.array[float]): Array of candidate scores
        """
        fixed_affiliation_string = fix_text(raw_affiliation).lower().replace(",", "")
        X = []
        for i, s in zip(candidates, scores):
            ror_entry = parse_ror_entry_into_single_string_lightgbm(i, self.ror_index)
            x = make_lightgbm_features(fixed_affiliation_string, ror_entry, self.lm)
            x[-3:] = [s, int(s == -0.15), int(s == -0.1)]
            X.append(x)
        X = np.array(X)
        scores = self.model.predict(X, num_threads=self.num_threads)
        # penalty when no match across fields
        has_no_match = X[:, self.inds_to_check].sum(1) == 0
        scores -= 0.05 * has_no_match  # magic number!
        scores += 0.05 * X[:, self.city_ind]
        scores_argsort = np.argsort(scores)[::-1]
        reranked = np.vstack([np.array(candidates), scores]).T[scores_argsort]
        return reranked[:, 0], reranked[:, 1].astype(float)



import re
import numpy as np
from nltk.util import ngrams
from s2aff.text import fix_text, STOPWORDS


def parse_ror_entry_into_single_string_lightgbm(ror_id_or_other_affiliation, ror_index):
    if ror_id_or_other_affiliation not in ror_index.ror_dict:
        # assuming that this is just not a ror_id and instead a string of some affiliation
        return {
            "names": [ror_id_or_other_affiliation],
            "acronyms": [],
            "city": [],
            "state": [],
            "country": [],
            "works_count": 0,
        }
    else:
        # otherwise it's a ROR id
        ror_id = ror_id_or_other_affiliation

    ror_entry = ror_index.ror_dict[ror_id]
    official_name = fix_text(ror_entry["name"]).lower().replace(",", "")
    aliases = [fix_text(i).lower().replace(",", "") for i in ror_entry["aliases"]]
    labels = [
        fix_text(i["label"]).lower().replace(",", "") for i in ror_entry["labels"]
    ]  # e.g. Sorbonne University has a label of Sorbonne Universités
    acronyms = [fix_text(i).lower().replace(",", "") for i in ror_entry["acronyms"]]

    output = {"names": list(set([official_name] + aliases + labels)), "acronyms": acronyms}

    if "addresses" in ror_entry and len(ror_entry["addresses"]) > 0:
        city = fix_text(ror_entry["addresses"][0]["city"] or "").lower().replace(",", "")

        #
        state = fix_text(ror_entry["addresses"][0]["state"] or "").lower().replace(",", "")

        if ror_entry["addresses"][0]["state_code"] is not None:
            state_code = ror_entry["addresses"][0]["state_code"].split("-")
        else:
            state_code = []

        # sometimes the state code is just a number - remove it if so
        state_code = [fix_text(i).lower().replace(",", "") for i in state_code if i.isalpha()]

        # country
        country_and_codes = ror_index.country_codes_dict[ror_entry["addresses"][0]["country_geonames_id"]]
        if country_and_codes[0] == "China":
            extras = ["PR", "PRC"]
        else:
            extras = []

        seen = set()
        country_elements = []
        for i in country_and_codes + extras:
            i_fix = fix_text(i).lower().replace(",", "")
            if i_fix not in seen and len(i_fix) > 0:
                country_elements.append(i_fix)
                seen.add(i_fix)

        state_info = [state] + state_code
        # sometimes the country code ends up being a state code
        state_info = [i for i in state_info if i not in seen]
        # sometimes the state is also a city
        state_info = [i for i in state_info if i != city]

        output["city"] = [city]
        output["state"] = state_info
        output["country"] = country_elements
    else:
        output["city"] = []
        output["state"] = []
        output["country"] = []

    output["works_count"] = ror_entry["works_count"]

    return output


def split(delimiters, string, maxsplit=0):
    regexPattern = "|".join(map(re.escape, delimiters))
    return re.split(regexPattern, string, maxsplit)


def find_query_ngrams_in_text(q, t, len_filter=1, remove_stopwords=True, use_word_boundaries=False, max_ngram_len=7):
    """A function to find instances of ngrams of query q
    inside text t. Finds all possible ngrams and returns their
    character-level span.

    Note: because of the greedy match this function can miss
    some matches when there's repetition in the query, but
    this is likely rare enough that we can ignore it

    Arguments:
        q {str} -- query
        t {str} -- text to find the query within
        len_filter {int} -- shortest allowable matches in characters
        remove_stopwords {bool} -- whether to remove stopwords-only matches
        use_word_boundaries {bool} -- whether to care about word boundaries
                                      when finding matches
        max_ngram_len {int} -- longest allowable derived word n-grams

    Returns:
        match_spans -- a list of span tuples
        match_text_tokenized -- a list of matched tokens
    """

    if len(q) == 0 or len(t) == 0:
        return [], []
    if type(q[0]) is not str or type(t) is not str:
        return [], []

    q_split = q.split()
    n_grams = []
    longest_ngram = np.minimum(max_ngram_len, len(q_split))
    for i in range(int(longest_ngram), 0, -1):
        n_grams += [" ".join(ngram).replace("|", "\|") for ngram in ngrams(q_split, i)]

    if use_word_boundaries:
        matches = list(re.finditer("|".join(["\\b" + i + "\\b" for i in n_grams]), t))
    else:
        matches = list(re.finditer("|".join(n_grams), t))
    match_spans = [i.span() for i in matches if i.span()[1] - i.span()[0] > len_filter]
    match_text_tokenized = [i.group() for i in matches if i.span()[1] - i.span()[0] > len_filter]

    # now we remove any of the results if the entire matched ngram is just a stopword
    if remove_stopwords:
        match_spans = [span for i, span in enumerate(match_spans) if match_text_tokenized[i] not in STOPWORDS]
        match_text_tokenized = [text for text in match_text_tokenized if text not in STOPWORDS]

    return match_spans, match_text_tokenized


def make_feature_names_and_constraints():
    # we get these counts from openalex and unclear how good they are
    # it works better to just take a ballpark value here
    # instead of the exact value.
    feats = ["works_count"]

    # for lightgbm, 1 means positively monotonic, -1 means negatively monotonic and 0 means non-constraint
    constraints = ["1"]

    # features for each field in ROR
    for field in ["names", "acronyms", "city", "state", "country"]:

        feats.extend(
            [
                f"{field}_frac_of_query_matched_in_text",  # total fraction of the query that was matched in text
                f"{field}_mean_of_log_probs",  # statistics of the log-probs
                f"{field}_sum_of_log_probs*match_lens",
                f"{field}_any_text_matched_in_query",
            ]
        )

        constraints.extend(
            [
                "1",
                "-1",
                "-1",
                "1",
            ]
        )

    feats.extend(
        [
            "fraction_of_query_matched_across_all_fields",
            "sum_log_prob_of_unmatched_unigrams",
            "stage_1_score",
            "stage_1_score_is_-0.15",
            "stage_1_score_is_-0.10",
        ]
    )

    constraints.extend(["1", "1", "0", "0", "0"])

    return np.array(feats), ",".join(constraints)


FEATURE_NAMES, FEATURE_CONSTRAINTS = make_feature_names_and_constraints()


def make_lightgbm_features(query, ror_entry, lm, max_q_len=256):
    def lm_score(s):
        return lm.score(s, eos=False, bos=False)

    # later we will filter some features based on nonsensical unigrams in the query
    # this is the log probability lower-bound for sensible unigrams
    log_prob_nonsense = lm_score("qwertyuiop")

    # fix and lowercase the text
    q = str(query)[:max_q_len]

    # if there's no query left at this point, we return NaNs
    # which the model natively supports
    if len(q) == 0:
        return [np.nan] * len(FEATURE_NAMES)

    q_split_set = set(q.split()) - STOPWORDS
    q_set_len = len(" ".join(q_split_set))
    q_set_len = np.maximum(q_set_len, 1)  # to avoid division by zero later

    matched_across_fields = []
    feats = [np.round(np.log1p(ror_entry["works_count"]))]
    names_and_acronyms_matches = set()
    for field in ["names", "acronyms", "city", "state", "country"]:

        match_spans = []
        match_text = []

        text_in_query = []
        for text in ror_entry[field]:
            # forward search
            ms, mt = find_query_ngrams_in_text(q, text)
            match_spans.extend(ms)
            match_text.extend(mt)
            matched_across_fields.extend(match_text)
            text_in_query.append(int(text in q))

            # reverse search
            ms, mt = find_query_ngrams_in_text(text, q)
            match_spans.extend(ms)
            match_text.extend(mt)
            matched_across_fields.extend(match_text)

        # take the set of the results while excluding sub-ngrams if longer ngrams are found
        # e.g. if we already matched 'sentiment analysis', then 'sentiment' is excluded
        match_spans_set = []
        match_text_set = []
        for t, s in sorted(zip(match_text, match_spans), key=lambda s: len(s[0]))[::-1]:
            if t not in match_text_set and ~np.any([t in i for i in match_text_set]):
                match_spans_set.append(s)
                match_text_set.append(t)

        # match_text_set but unigrams
        matched_text_unigrams = set()
        for i in match_text_set:
            i_split = i.split()
            matched_text_unigrams.update(i_split)
            if field in {"names", "acronyms"}:
                names_and_acronyms_matches.update(i_split)

        if len(match_text_set) > 0:  # if any matches
            # log probabilities of the scores
            lm_probs = [lm_score(match) for match in match_text_set]

            match_word_lens = [len(i.split()) for i in match_text_set]

            # remove stopwords from unigrams
            matched_text_unigrams -= STOPWORDS
            matched_text_len = len(" ".join(matched_text_unigrams))

            feats.extend(
                [
                    # total fraction of the query that was matched in text
                    matched_text_len / q_set_len,
                    np.nanmean(lm_probs),  # average log-prob of the matches
                    # sum of log-prob of matches times word-lengths
                    np.nansum(np.array(lm_probs) * np.array(match_word_lens)),
                    np.any(text_in_query),
                ]
            )
        else:
            # if we have no matches, then the features are deterministically 0
            feats.extend([0, 0, 0, 0])

    # special features for how much of the query was matched/unmatched across all fields
    if len(q_split_set) > 0:
        matched_split_set = set()
        for i in matched_across_fields:
            matched_split_set.update(i.split())
        # making sure stopwords aren't an issue
        matched_split_set -= STOPWORDS
        # fraction of the query matched
        matched_len = len(" ".join(matched_split_set))
        feats.append(matched_len / q_set_len)
        # the log-prob of the unmatched quotes
        unmatched = q_split_set - matched_split_set
        log_probs_unmatched_unquoted = [lm_score(i) for i in unmatched]
        feats.append(np.nansum([i for i in log_probs_unmatched_unquoted if i > log_prob_nonsense]))
        feats.extend([np.nan] * 3)  # this will later be populated by the first stage ranker score
    else:
        feats.extend([np.nan] * 5)

    return feats



"""
How long does this all take?

Times below are all run on s2-server2 on the CPU. 
"""
from time import time
import os, psutil
from s2aff.consts import PATHS
from s2aff.ror import RORIndex
from s2aff.model import NERPredictor, PairwiseRORLightGBMReranker, parse_ner_prediction
from s2aff.data import load_gold_affiliation_annotations
from s2aff.text import fix_text
from blingfire import text_to_words

USE_CUDA = False

print(
    "Total memory footprint before loading various artifacts (in MB): ",
    psutil.Process(os.getpid()).memory_info().rss / 1024**2,
)

# fixed costs: load everything
ner_predictor = NERPredictor(use_cuda=USE_CUDA)
print(
    "Total memory footprint after loading NER (in MB): ",
    psutil.Process(os.getpid()).memory_info().rss / 1024**2,
)
ror_index = RORIndex()
print(
    "Total memory footprint after loading inverted index (in MB): ",
    psutil.Process(os.getpid()).memory_info().rss / 1024**2,
)
print(
    "Total memory footprint after loading lightgbm (in MB): ",
    psutil.Process(os.getpid()).memory_info().rss / 1024**2,
)
pairwise_model = PairwiseRORLightGBMReranker(ror_index, num_threads=0)  # 0 is default open threads

# what is the total memory footprint at this point?
print(
    "Total memory footprint after loading everything (in MB): ",
    psutil.Process(os.getpid()).memory_info().rss / 1024**2,
)


# some data to test on - 100 samples
df = load_gold_affiliation_annotations()
texts = df.original_affiliation.values[-100:].tolist()

# prepare for NER
# texts = [text_to_words(fix_text(text)) for text in texts]

print("Longest test affiliation (# characters):", max(map(len, texts)))
print("Shortest test affiliation (# characters):", min(map(len, texts)))

# NER time: 9.02 seconds on CPU
# there are no batch size efficiency gains above 8

start = time()
preds = ner_predictor.predict(texts)
end = time()
print(f"NER on CPU: {end - start}")

# parsing is negligible: 0.08 seconds
parsed_tuples = []
start = time()
for i in preds:
    main, child, address, early_candidates = parse_ner_prediction(i, ror_index)
    parsed_tuples.append((main, child, address, early_candidates))
end = time()
print(f"Parsing NER Preds: {end - start}")

# stage 1 ranker time: 27s
candidates_and_scores = []
start = time()
for main, child, address, early_candidates in parsed_tuples:
    candidates, scores = ror_index.get_candidates_from_main_affiliation(main, address, early_candidates)
    candidates_and_scores.append((candidates, scores))
end = time()
print(f"Stage 1 Candidates: {end - start}")

# lightgbm: 12.6s for top 100
raw_affiliations = df.original_affiliation.values[-100:].tolist()
start = time()
for raw_affiliation, (candidates, scores) in zip(raw_affiliations, candidates_and_scores):
    reranked_candidates, reranked_scores = pairwise_model.predict(raw_affiliation, candidates[:100], scores[:100])
end = time()
print(f"LightGBM: {end - start}")

print(
    "Total memory footprint after running everything (in MB): ",
    psutil.Process(os.getpid()).memory_info().rss / 1024**2,
)



import re
from text_unidecode import unidecode
import pandas as pd

RE_BASIC = re.compile(r"[^,a-zA-Z0-9\s]+")
RE_APOSTRAPHE_S = re.compile(r"(\w+)'s")

ABBREVIATION_DICTIONARY = {
    "Academy": "Acad",
    "American": "Amer",
    "Association": "Assoc",
    "College": "Coll",
    "Company": "Co",
    "Corporation": "Corp",
    "Communication": "Commun",
    "Department": "Dept",
    "Division": "Div",
    "Doctor": "Dr",
    "Electrical": "Electr",
    "Engineering": "Eng",
    "European": "Europ",
    "Executive": "Exec",
    "Faculty": "Fac",
    "Foundation": "Found",
    "Government": "Gov",
    "Incorporated": "Inc",
    "Information": "Info",
    "Institute": "Inst",
    "International": "Intern",
    "Laboratory": "Lab",
    "Library": "Libr",
    "National": "Nat",
    "Medicine": "Med",
    "Mechanical": "Mech",
    "Professor": "Prof",
    "Program": "Progr",
    "Psychology": "Psychol",
    "School": "Sch",
    "Science": "Sci",
    "Society": "Soc",
    "Technology": "Technol",
    "University": "Univ",
}

INVERTED_ABBREVIATION_DICTIONARY = {v.lower(): k.lower() for k, v in ABBREVIATION_DICTIONARY.items()}
INVERTED_ABBREVIATION_DICTIONARY["tech"] = "technology"

STOPWORDS = set(
    [
        "it",
        "its",
        "is",
        "are",
        "a",
        "an",
        "the",
        "and",
        "as",
        "of",
        "at",
        "by",
        "for",
        "with",
        "into",
        "from",
        "in",
    ]
)


CERTAINLY_MAIN = {
    "Associates",
    "Trust",
    "Foundation",
    "Society",
    "Museum",
    "Association",
    "Universitat",
    "Univ",
    "Universita",
    "Universite",
    "Universidad",
    "Universidade",
    "Universitaria",
    "University",
    "Universitas",
    "Institute",
    "Institut",
    "Center",
    "Centre",
    "Centro",
    "Istituto",
}


def fix_text(s):
    """General purpose text fixing"""
    if pd.isnull(s) or len(s) == 0:
        return ""

    s = unidecode(s).replace("#TAB#", "").replace(".", "").replace(" & ", " and ").replace("&", "n")
    s = RE_APOSTRAPHE_S.sub(r"\1s", s)
    s = RE_BASIC.sub(" ", s)
    s = re.sub(r"\s+", " ", s).strip()
    return s



import os

os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
os.environ["CUDA_VISIBLE_DEVICES"] = "0"  # change to your reserved GPU
import pickle
import kenlm
from random import sample
import numpy as np
from s2aff.text import fix_text
from s2aff.consts import PATHS
from s2aff.ror import RORIndex
from s2aff.features import make_lightgbm_features, parse_ror_entry_into_single_string_lightgbm
from s2aff.model import NERPredictor, parse_ner_prediction
from s2aff.data import load_gold_affiliation_annotations


lm = kenlm.LanguageModel(PATHS["kenlm_model"])
ror_index = RORIndex()


def make_pairwise_data(df_gold, lm, ror_index, split_str="train", n_negatives=100, pos_weight=10, use_alt_pos=True):
    X = []
    y = []
    w = []
    rors = []
    qs = []
    group = []

    def make_single_pair(text, ror_id, label, weight, ind, score, suffix=""):
        x = make_lightgbm_features(text, parse_ror_entry_into_single_string_lightgbm(ror_id, ror_index), lm)
        x[-3:] = [score, int(score == -0.15), int(score == -0.1)]
        X.append(x)
        y.append(label)
        w.append(weight)
        group.append(ind)
        rors.append(ror_id)
        qs.append(text + suffix)

    outer_ind = 0
    if split_str is not None:
        df_sub = df_gold[df_gold.split == split_str]
    else:
        df_sub = df_gold
    for _, row in df_sub.iterrows():
        main, child, address, early_candidates = parse_ner_prediction(row.parsed_affiliations, ror_index)
        candidates, scores = ror_index.get_candidates_from_main_affiliation(main, address, early_candidates)
        gold = set(row.labels)
        candidates_positives_dict = {c: s for c, s in zip(candidates, scores)}
        candidates_positives = [(i, candidates_positives_dict.get(i, 0.001)) for i in gold]
        # top 100 candidates are what we will rerank in prod, so sampling those here
        candidates_negatives = [(i, s) for i, s in zip(candidates[:n_negatives], scores[:n_negatives]) if i not in gold]
        if len(candidates_negatives) > 0:
            candidates_negatives, scores_negatives = zip(*candidates_negatives)
        else:
            candidates_negatives, scores_negatives = [], []
        original_affiliation_fixed = fix_text(row.original_affiliation).lower().replace(",", "")
        main = fix_text(" ".join(main)).lower().replace(",", "")
        address = fix_text(" ".join(address)).lower().replace(",", "")

        for gold_ror_id, positive_score in candidates_positives:
            make_single_pair(original_affiliation_fixed, gold_ror_id, 1, pos_weight, outer_ind, positive_score)

        for neg_ror_id, negative_score in zip(candidates_negatives, scores_negatives):
            make_single_pair(original_affiliation_fixed, neg_ror_id, 0, 1, outer_ind, negative_score)

        outer_ind += 1

        if use_alt_pos:

            if main != original_affiliation_fixed:

                for gold_ror_id, positive_score in candidates_positives:
                    make_single_pair(main, gold_ror_id, 1, pos_weight, outer_ind, positive_score, suffix=" [MAIN ONLY]")

                for neg_ror_id, negative_score in zip(candidates_negatives, scores_negatives):
                    make_single_pair(main, neg_ror_id, 0, 1, outer_ind, negative_score, suffix=" [MAIN ONLY]")

                outer_ind += 1

            if address != "" and (main + " " + address) != original_affiliation_fixed:
                for gold_ror_id, positive_score in candidates_positives:
                    make_single_pair(
                        main + " " + address,
                        gold_ror_id,
                        1,
                        pos_weight,
                        outer_ind,
                        positive_score,
                        suffix=" [MAIN+ADDRESS]",
                    )

                for neg_ror_id, negative_score in zip(candidates_negatives, scores_negatives):
                    make_single_pair(
                        main + " " + address, neg_ror_id, 0, 1, outer_ind, negative_score, suffix=" [MAIN+ADDRESS]"
                    )

                outer_ind += 1

    return np.array(X), np.array(y), np.array(w), np.array(group), np.array(rors), np.array(qs)


"""
Training set for gold data
"""
df_gold = load_gold_affiliation_annotations()

# should do this batched
ner_predictor = NERPredictor(use_cuda=False)
parsed_affiliations = ner_predictor.predict(df_gold.loc[:, "original_affiliation"].values)
df_gold.loc[:, "parsed_affiliations"] = parsed_affiliations
ner_predictor.delete_model()  # clean up to make space

X_train, y_train, w_train, group_train, rors_train, qs_train = make_pairwise_data(
    df_gold,
    lm,
    ror_index,
    split_str="train",
    n_negatives=100,
    use_alt_pos=True,
    pos_weight=10,
)
X_val, y_val, w_val, group_val, rors_val, qs_val = make_pairwise_data(
    df_gold,
    lm,
    ror_index,
    split_str="val",
    n_negatives=100,
    use_alt_pos=False,
    pos_weight=10,
)
X_test, y_test, w_test, group_test, rors_test, qs_test = make_pairwise_data(
    df_gold,
    lm,
    ror_index,
    split_str="test",
    n_negatives=100,
    use_alt_pos=False,
    pos_weight=10,
)

with open(PATHS["lightgbm_gold_features"], "wb") as f:
    pickle.dump(
        [
            (X_train, y_train, w_train, group_train, rors_train, qs_train),
            (X_val, y_val, w_val, group_val, rors_val, qs_val),
            (X_test, y_test, w_test, group_test, rors_test, qs_test),
        ],
        f,
    )

# same but do it on the stuff with no ROR at all
df_gold = load_gold_affiliation_annotations(keep_non_ror_gold=True)
keep_flag = df_gold.labels.apply(lambda x: any(["ror.org" not in i for i in x]))
df_gold = df_gold[keep_flag]

# should do this batched
ner_predictor = NERPredictor(use_cuda=False)
parsed_affiliations = ner_predictor.predict(df_gold.loc[:, "original_affiliation"].values)
df_gold.loc[:, "parsed_affiliations"] = parsed_affiliations
ner_predictor.delete_model()  # clean up to make space

X_none, y_none, w_none, group_none, rors_none, qs_none = make_pairwise_data(
    df_gold,
    lm,
    ror_index,
    split_str=None,
    n_negatives=100,
    use_alt_pos=False,
    pos_weight=10,
)

with open(PATHS["lightgbm_gold_no_ror_features"], "wb") as f:
    pickle.dump(
        (X_none, y_none, w_none, group_none, rors_none, qs_none),
        f,
    )



import logging
from s2aff.model import parse_ner_prediction

logger = logging.getLogger("s2aff")
logger.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
ch = logging.StreamHandler()
ch.setFormatter(formatter)
ch.setLevel(logging.INFO)
logger.addHandler(ch)


class S2AFF:
    """
    The wrapper class that links a raw affiliation string to a ROR entry.

    :param ner_predictor: an instantiated NERPredictor object
    :param ror_index: an instantiated RORIndex object
    :param pairwise_model: an instantiated pairwise model with `predict` method that
        returns a list of reranked candidates and a list of reranked scores
        best we have is `PairwiseRORLightGBMReranker`
    :param top_k_first_stage: the number of top first stage RORINdex candidates to rerank
        default = 100
    :param pairwise_model_threshold: the threshold for the pairwise model,
        if the top score is below this threshold, and the difference between the top
        two scores is below the delta threshold, then there is no ROR match
        default = 0.3, found in `scripts/analyze_thresholds.ipynb`
    :param pairwise_model_delta_threshold: the delta threshold for the pairwise model
        if the difference between the top two scores is below this threshold and
        the top score is below the threshold (previous param) then there is no ROR match
        default = 0.2, found in `scripts/analyze_thresholds.ipynb`
    :param no_ror_output_text: the text to return if no ROR is found due to thereshold
        default = "NO_ROR_FOUND"
    :param no_candidates_output_text: the text to return if no candidates are found in the first stage
        default = "NO_CANDIDATES_FOUND"
    :param number_of_top_candidates_to_return: the number of top candidates to return in the second stage of the algorithm
        a convenience to reduce the total amount of data sent
        default = 5
    :param look_for_grid_and_isni: whether to look for GRID and ISNI ids in the raw affiliation string
        ff found, just returns that candidate as the only one to the subsequent step
        default = True
    """

    def __init__(
        self,
        ner_predictor,
        ror_index,
        pairwise_model,
        top_k_first_stage=100,
        pairwise_model_threshold=0.3,
        pairwise_model_delta_threshold=0.2,
        no_ror_output_text="NO_ROR_FOUND",
        no_candidates_output_text="NO_CANDIDATES_FOUND",
        number_of_top_candidates_to_return=5,
        look_for_grid_and_isni=True,
    ):
        self.ner_predictor = ner_predictor
        self.ror_index = ror_index
        self.pairwise_model = pairwise_model
        self.top_k_first_stage = top_k_first_stage
        self.pairwise_model_threshold = pairwise_model_threshold
        self.pairwise_model_delta_threshold = pairwise_model_delta_threshold
        self.no_ror_output_text = no_ror_output_text
        self.no_candidates_output_text = no_candidates_output_text
        self.number_of_top_candidates_to_return = number_of_top_candidates_to_return
        self.look_for_grid_and_isni = True

    def predict(self, raw_affiliations):
        """Predict function for raw affiliation strings

        :param raw_affiliations: a list of raw affiliation strings

        :return: a list of dictionaries with the following keys:
            "raw_affiliation": raw affiliation string
            "ner_prediction": raw NER prediction
            "main_from_ner": main institute portion of the raw_affiliation extracted from NER prediction
            "child_from_ner": child institute (if any) portion of the raw_affiliation extracted from NER prediction
            "address_from_ner": address portion (if any) of the raw_affiliation extracted from NER prediction
            "stage1_candidates": top candidates from the RORIndex
            "stage1_scores": scores of the above candidates
            "stage2_candidates": top candidates from the pairwise model
            "stage2_scores": scores of the above candidates
        """
        # if we get a single string, we should make a list of out of it
        if isinstance(raw_affiliations, str):
            raw_affiliations = [raw_affiliations]

        print("Getting NER predictions in bulk...")
        ner_predictions = self.ner_predictor.predict(raw_affiliations)
        print("Done")

        outputs = []
        for counter, (raw_affiliation, ner_prediction) in enumerate(zip(raw_affiliations, ner_predictions)):
            print(
                f"Getting ROR candidates and reranking for: '{raw_affiliation}' ({counter+1}/{len(raw_affiliations)})",
                end="\r",
            )
            main, child, address, early_candidates = parse_ner_prediction(ner_prediction, self.ror_index)
            # sometimes the affiliation strings just contain GRID or ISNI ids
            # todo: some time in the future the strings may contain ROR ids too
            if self.look_for_grid_and_isni:
                ror_from_grid = self.ror_index.extract_grid_and_map_to_ror(raw_affiliation)
                ror_from_isni = self.ror_index.extract_isni_and_map_to_ror(raw_affiliation)
                ror_from_grid_or_isni = ror_from_grid or ror_from_isni
                found_early = ror_from_grid_or_isni is not None
                if found_early:
                    candidates, scores = [ror_from_grid_or_isni], [1.0]
            else:
                found_early = False
            # we don't want to rerank if we found a GRID or ISNI id
            if not found_early:
                candidates, scores = self.ror_index.get_candidates_from_main_affiliation(
                    main, address, early_candidates
                )

            if len(candidates) == 0:
                output_scores_and_thresh = [self.no_candidates_output_text], [0.0]
            else:
                reranked_candidates, reranked_scores = self.pairwise_model.predict(
                    raw_affiliation, candidates[: self.top_k_first_stage], scores[: self.top_k_first_stage]
                )
                # apply threshold to reranked scores
                if len(reranked_candidates) == 0:
                    output_scores_and_thresh = [self.no_candidates_output_text], [0.0]
                elif len(reranked_candidates) == 1:
                    if reranked_scores[0] < self.pairwise_model_threshold:
                        output_scores_and_thresh = [self.no_ror_output_text], [0.0]
                    else:
                        output_scores_and_thresh = (reranked_candidates, reranked_scores)
                else:
                    delta = reranked_scores[0] - reranked_scores[1]
                    if (
                        reranked_scores[0] < self.pairwise_model_threshold
                        and delta < self.pairwise_model_delta_threshold
                    ):
                        output_scores_and_thresh = [self.no_ror_output_text], [0.0]
                    else:
                        output_scores_and_thresh = (reranked_candidates, reranked_scores)
            try:
                display_name = self.ror_index.ror_dict[output_scores_and_thresh[0][0]]["name"]
            except:
                display_name = ""

            # make a dict of outputs
            output = {
                "raw_affiliation": raw_affiliation,
                "ner_prediction": ner_prediction,
                "main_from_ner": main,
                "child_from_ner": child,
                "address_from_ner": address,
                "stage1_candidates": list(candidates[: self.number_of_top_candidates_to_return]),
                "stage1_scores": list(scores[: self.number_of_top_candidates_to_return]),
                "stage2_candidates": list(output_scores_and_thresh[0][: self.number_of_top_candidates_to_return]),
                "stage2_scores": list(output_scores_and_thresh[1][: self.number_of_top_candidates_to_return]),
                "top_candidate_display_name": display_name,
            }

            outputs.append(output)
        return outputs



# -*- coding: utf-8 -*-
"""
Created on Tue May 29 12:35:14 2018
@author: serge
"""

import time
from abc import ABCMeta, abstractmethod
from itertools import groupby
import numpy as np
import lightgbm as lgb
from hyperopt import hp, fmin, tpe, Trials, STATUS_OK, STATUS_FAIL


def format_group(g):
    return np.array([len(list(group)) for _, group in groupby(g)])


class Experiment(object, metaclass=ABCMeta):

    # This method must be overridden but can still be called via super by subclasses have shared construction logic
    @abstractmethod
    def __init__(self):
        pass

    def fit(self, params, dtrain, dtest, n_estimators, seed=0):
        raise NotImplementedError("Method fit is not implemented.")

    def convert_data(self, X, y, group):
        raise NotImplementedError("Method fit is not implemented.")

    def run_optimization(self, train_dmatrix, val_dmatrix, params, n_estimators=None, verbose=False):
        n_estimators = n_estimators or self.n_estimators
        best_n_estimators = n_estimators
        evals_results, start_time = [], time.time()
        _, evals_result = self.fit(params, train_dmatrix, val_dmatrix, n_estimators, self.random_seed)
        try:
            evals_result_final = evals_result[best_n_estimators - 1]
        except:
            evals_result_final = evals_result

        eval_time = time.time() - start_time

        cv_result = {
            "loss": self.eval_multiplier * evals_result_final,
            "best_n_estimators": best_n_estimators,
            "eval_time": eval_time,
            "status": STATUS_FAIL if np.isnan(evals_result_final) else STATUS_OK,
            "params": params.copy(),
        }
        self.best_loss = min(self.best_loss, cv_result["loss"])
        self.hyperopt_eval_num += 1
        cv_result.update({"hyperopt_eval_num": self.hyperopt_eval_num, "best_loss": self.best_loss})

        if verbose:
            print(
                "[{0}/{1}]\teval_time={2:.2f} sec\tcurrent_{3}={4:.6f}\tmin_{3}={5:.6f}".format(
                    self.hyperopt_eval_num,
                    self.hyperopt_evals,
                    eval_time,
                    self.eval_metric,
                    cv_result["loss"],
                    self.best_loss,
                )
            )
        return cv_result

    def run_test(self, dtrain, dtest, params=None, n_estimators=None, custom_metric=None, seed=0):
        params = params or self.best_params
        n_estimators = n_estimators or self.best_n_estimators or self.n_estimators
        start_time = time.time()
        self.bst, evals_result = self.fit(params, dtrain, dtest, n_estimators, seed=seed)
        try:
            evals_result_final = evals_result[-1]
        except:
            evals_result_final = evals_result
        eval_time = time.time() - start_time
        result = {
            "loss": evals_result_final,
            "n_estimators": n_estimators,
            "eval_time": eval_time,
            "status": STATUS_OK,
            "params": params.copy(),
        }
        return result

    def run(
        self,
        X_train,
        y_train,
        group_train,
        weights_train,
        X_val,
        y_val,
        group_val,
        weights_val,
        X_test,
        y_test,
        group_test,
        weights_test,
    ):
        print("Loading and preprocessing dataset...")
        train_dmatrix = self.convert_data(X_train, y_train, group_train, weights_train)
        val_dmatrix = self.convert_data(X_val, y_val, group_val, weights_val)
        test_dmatrix = self.convert_data(X_test, y_test, group_test, weights_test)

        print("Optimizing params...")
        cv_result = self.optimize_params(train_dmatrix, val_dmatrix)
        self.print_result(cv_result, "\nBest result on cv")

        print("\nTraining algorithm with the tuned parameters for different seeds...")
        test_losses = []
        for seed in [42, 1999, self.random_seed]:
            test_result = self.run_test(train_dmatrix, test_dmatrix, seed=seed)
            test_losses.append(test_result["loss"])
            print("For seed=%d Test's %s : %.5f" % (seed, self.eval_metric, test_losses[-1]))

        print(
            "\nTest's %s mean: %.5f, Test's %s std: %.5f"
            % (self.eval_metric, np.mean(test_losses), self.eval_metric, np.std(test_losses))
        )

        return test_losses

    def optimize_params(self, train_dmatrix, val_dmatrix, max_evals=None, verbose=True):
        max_evals = max_evals or self.hyperopt_evals
        self.trials = Trials()
        self.hyperopt_eval_num, self.best_loss = 0, np.inf

        fmin(
            fn=lambda params: self.run_optimization(train_dmatrix, val_dmatrix, params, verbose=verbose),
            space=self.space,
            algo=tpe.suggest,
            max_evals=max_evals,
            trials=self.trials,
        )

        self.best_params = self.trials.best_trial["result"]["params"]
        if "best_n_estimators" in self.trials.best_trial["result"].keys():
            self.best_n_estimators = self.trials.best_trial["result"]["best_n_estimators"]
        return self.trials.best_trial["result"]

    def print_result(self, result, name="", extra_keys=None):
        print("%s:\n" % name)
        print("%s = %s" % (self.eval_metric, result["loss"]))
        if "best_n_estimators" in result.keys():
            print("best_n_estimators = %s" % result["best_n_estimators"])
        elif "n_estimators" in result.keys():
            print("n_estimators = %s" % result["n_estimators"])
        print("params = %s" % result["params"])
        if extra_keys is not None:
            for k in extra_keys:
                if k in result:
                    print("%s = %f" % (k, result[k]))


class lightgbmExperiment(Experiment):
    def __init__(
        self,
        learning_task,
        metric,
        eval_metric=None,
        label_gain_max=1000,
        eval_at="1,2,3,4,5,100",
        n_estimators=5000,
        hyperopt_evals=25,
        device="cpu",
        tree_learner="data",
        random_seed=0,
        threads=16,
        feval=None,
    ):

        super().__init__()

        if learning_task == "classification":
            assert metric in {"binary", "multiclass", "multiclassova"}
        elif learning_task == "regression":
            assert metric in {"rmse", "mae", "huber", "fair", "poisson", "quantile", "mape", "gamma", "tweedie"}
        elif learning_task == "ranking":
            assert metric in {"lambdarank", "rank_xendcg"}

        self.learning_task = learning_task
        self.metric = metric
        self.n_estimators = n_estimators
        self.label_gain_max = int(label_gain_max) + 1
        self.hyperopt_evals = hyperopt_evals
        self.device = device
        self.random_seed = random_seed
        self.threads = threads
        self.feval = feval

        self.best_loss = np.inf
        self.best_n_estimators = None
        self.hyperopt_eval_num = 0

        # eval metric
        if eval_metric in {
            "auc",
            "binary_logloss",
            "binary_error",
            "multi_logloss",
            "multi_error",
            "l1",
            "l2",
            "rmse",
            "quantile",
            "mape",
            "huber",
            "fair",
            "poisson",
            "gamma",
            "gamma_deviance",
            "tweedie",
            "ndcg",
            "map",
        }:
            self.eval_metric = eval_metric
        else:
            self.eval_metric = metric  # same as objective

        self.space = {
            "learning_rate": hp.loguniform("learning_rate", -7, 0),
            "num_leaves": hp.qloguniform("num_leaves", 2, 7, 1),
            "feature_fraction": hp.uniform("feature_fraction", 0.5, 1),
            "bagging_fraction": hp.uniform("bagging_fraction", 0.5, 1),
            "min_data_in_leaf": hp.qloguniform("min_data_in_leaf", 2, 7, 1),
            "min_sum_hessian_in_leaf": hp.loguniform("min_sum_hessian_in_leaf", -16, 5),
            "lambda_l1": hp.loguniform("lambda_l1", -16, 2),
            "lambda_l2": hp.loguniform("lambda_l2", -16, 2),
        }

        self.constant_params = {
            "max_bin": 63,
            "device": self.device,
            "gpu_use_dp": False,
            "min_gain_to_split": 0.1,  # try to control overfitting
            "num_threads": self.threads,
            "tree_learner": tree_learner,
            "objective": self.metric,
            "metric": self.eval_metric,
            "label_gain": ",".join([str(int(np.log2(i + 1))) for i in range(self.label_gain_max)]),
            "bagging_freq": 1,
            "verbose": -1,
            "eval_at": eval_at,
        }

        # hyperopt minimizes metrics, so we have negate ones where larger is better
        if self.eval_metric in {"auc", "ndcg", "map"}:
            self.eval_multiplier = -1
        else:
            self.eval_multiplier = 1

    def fit(self, params, dtrain, dtest, n_estimators, seed=0):
        params["num_leaves"] = max(int(params["num_leaves"]), 2)
        params["min_data_in_leaf"] = int(params["min_data_in_leaf"])
        params.update(self.constant_params)
        params.update(
            {
                "feature_fraction_seed": 2 + seed,
                "bagging_seed": 3 + seed,
                "drop_seed": 4 + seed,
            }
        )
        if self.metric in {"multiclass", "multiclassova"} and "num_class" not in params:
            params["num_class"] = len(np.unique(dtrain.get_label()))

        evals_result = {}
        bst = lgb.train(
            params,
            dtrain,
            valid_sets=dtest,
            valid_names=["test"],
            evals_result=evals_result,
            num_boost_round=n_estimators,
        )

        if self.eval_metric in {"ndcg", "map"}:
            key = f"{self.eval_metric}@1"
        else:
            key = self.eval_metric
        results = evals_result["test"][key]

        return bst, results

    def convert_data(self, X, y, group=None, weight=None, cat_cols=None):
        if cat_cols is None:
            categorical_feature = None
        else:
            categorical_feature = np.nonzero(cat_cols)[0].tolist()
        if len(group) == len(y):
            group = format_group(group)
        dataset = lgb.Dataset(
            X, y, group=group, weight=weight, categorical_feature=categorical_feature, free_raw_data=False
        )
        return dataset



"""
Our ROR file needs to have the works count from OpenAlex added to it.

This does that
"""

from s2aff.consts import PATHS
import json
import boto3
import gzip
from io import BytesIO
import pandas as pd


"""
Go through every single file from s3://openalex/data/institutions/updated_date=*/part-***.gz

Stream the jsonl files inside there
"""
bucket = "openalex"
prefix = "data/institutions"

s3 = boto3.client("s3")
paginator = s3.get_paginator("list_objects_v2")
pages = paginator.paginate(Bucket=bucket, Prefix=prefix)

works_count_dict = {}
for page in pages:
    for obj in page["Contents"]:
        if obj["Key"].startswith("part") and obj["Key"].endswith(".gz"):
            print("Working on", obj["Key"])
            obj = s3.get_object(Bucket=bucket, Key=obj["Key"])
            with gzip.GzipFile(fileobj=BytesIO(obj["Body"].read())) as f:
                for line in f:
                    line = json.loads(line)
                    ror_id = line["ror"]
                    works_count = line["works_count"]
                    works_count_dict[ror_id] = works_count

# convert works_count_dict to a dataframe
df = pd.DataFrame.from_dict(works_count_dict, orient="index", columns=["works_count"]).reset_index()
df.columns = ["ror", "works_count"]
df.to_csv(PATHS["openalex_works_counts"], index=False)



import pandas as pd
from s2aff.consts import PATHS
from s2aff.file_cache import cached_path


def load_gold_affiliation_annotations(gold_path=PATHS["gold_affiliation_annotations"], keep_non_ror_gold=False):
    """Load the gold annotation file.

    Args:
        gold_path (str, optional): Location of gold. Defaults to PATHS["gold_affiliation_annotations"].
        keep_non_ror_gold (bool, optional): Whether to keep annotations for rows without ROR entries.
            Defaults to False.

    Returns:
        gold_df: DataFrame with gold annotations.
    """
    df = pd.read_csv(cached_path(gold_path))
    df.loc[:, "labels"] = df["labels"].apply(eval)
    if not keep_non_ror_gold:
        keep_flag = df.labels.apply(lambda x: any(["ror.org" in i for i in x]))
        df = df[keep_flag]
    return df



import pickle
import numpy as np
import sklearn
import lightgbm as lgb
from s2aff.ror import RORIndex
from s2aff.text import fix_text
from s2aff.consts import PATHS
from s2aff.lightgbm_helpers import lightgbmExperiment
from s2aff.features import FEATURE_CONSTRAINTS, FEATURE_NAMES

FEATURE_NAMES = list(FEATURE_NAMES)

inds_to_check = [
    FEATURE_NAMES.index("names_frac_of_query_matched_in_text"),
    FEATURE_NAMES.index("acronyms_frac_of_query_matched_in_text"),
]

city_ind = FEATURE_NAMES.index("city_frac_of_query_matched_in_text")
stage_1_ind = FEATURE_NAMES.index("stage_1_score")


def eval_lgb(y, X, scores, group, weight=0.05):
    has_no_match = X[:, inds_to_check].sum(1) == 0
    scores_adjusted = scores - weight * has_no_match
    scores_adjusted = scores_adjusted + weight * X[:, city_ind]
    APs = []
    RR = []
    Pat1 = []
    sorted_row_inds = np.sort(list(set(group)))
    for i in sorted_row_inds:
        flag = group == i
        p = scores_adjusted[flag]
        y_true = y[flag]
        APs.append(sklearn.metrics.average_precision_score(y_true, p))
        argsort = np.argsort(p)[::-1]  # largest to smallest
        y_true_argsort = y_true[argsort]  # labels are sorted according to probs
        first_index_of_label_1 = list(y_true_argsort).index(1) + 1  # where is the first appearance of label 1?
        RR.append(1.0 / first_index_of_label_1)  # reciprocal of that
        Pat1.append(first_index_of_label_1 == 1)  # is it the first position?

    results = {
        "MAP": np.mean(APs),
        "MRR": np.mean(RR),
        "Prec@1": np.mean(Pat1),
        "MAP_vec": APs,
        "MRR_vec": RR,
        "Prec@1_vec": Pat1,
        "group_vec": sorted_row_inds,
    }
    print({"MAP": np.mean(APs), "MRR": np.mean(RR), "Prec@1": np.mean(Pat1)})
    return results


"""
Step 1: get all available raw affiliations and their grid ids, map to ROR, and save to a file.
We will also save a raw file with all the lower-cased affiliations only. 

Step 2: Run kenlm in the command line like this:
~/kenlm/build/bin/lmplz -o 6 -S 100G --skip_symbols --prune 0 3 -T /tmp <data/raw_affiliations_lowercased.txt >data/raw_affiliations_lowercased.arma
~/kenlm/build/bin/build_binary -T /tmp data/raw_affiliations_lowercased.arma data/raw_affiliations_lowercased.binary

These two commands were run offline, and the steps are not reproducible in this repo
"""


"""
Step 3: LightGBM ranker requires featurization for pairs.
We can mostly use the featurization from the search project,
but there are some differences:
- don't need year, citations, oldness, special treatment for authors, special treatment for quotes
- some fields have multiple entries, so need to combine those somehow
"""

with open(PATHS["lightgbm_gold_features"], "rb") as f:
    [
        (X_train, y_train, w_train, group_train, rors_train, qs_train),
        (X_val, y_val, w_val, group_val, rors_val, qs_val),
        (X_test, y_test, w_test, group_test, rors_test, qs_test),
    ] = pickle.load(f)

train_main_flag = np.array([i.endswith("[MAIN ONLY]") for i in qs_train])
train_main_address_flag = np.array([i.endswith("[MAIN+ADDRESS]") for i in qs_train])
train_augmented_flag = train_main_flag | train_main_address_flag

w_train = w_train.astype(np.float32)
w_train[train_main_flag] = w_train[train_main_flag] / 10
w_train[train_main_address_flag] = w_train[train_main_address_flag] / 2

lgb_exp = lightgbmExperiment(
    learning_task="ranking",
    metric="rank_xendcg",
    eval_metric="map",
    n_estimators=500,
    hyperopt_evals=50,
    device="cpu",
    eval_at="1,5,25,100",
    tree_learner="data",
    random_seed=7,
    threads=64,
)

# lgb_exp.constant_params["monotone_constraints"] = FEATURE_CONSTRAINTS  # generally makes things worse
lgb_exp.constant_params["feature_pre_filter"] = False  # prevents some obscure errors

test_ndcgs = lgb_exp.run(
    X_train,
    y_train,
    group_train,
    w_train,
    X_val,
    y_val,
    group_val,
    w_val,
    X_test,
    y_test,
    group_test,
    w_test,
)

# load the best model seen so far
# lgb_exp.bst = lgb.Booster(model_file=PATHS["lightgbm_model"])

# train_dmatrix = lgb_exp.convert_data(X_train, y_train, group_train, w_train)
# val_dmatrix = lgb_exp.convert_data(X_val, y_val, group_val, w_val)
# lgb_exp.bst, _ = lgb_exp.fit(lgb_exp.best_params, train_dmatrix, val_dmatrix, 500, 8)

train_results = eval_lgb(y_train, X_train, lgb_exp.bst.predict(X_train), group_train)
val_results = eval_lgb(y_val, X_val, lgb_exp.bst.predict(X_val), group_val)
test_results = eval_lgb(y_test, X_test, lgb_exp.bst.predict(X_test), group_test)

# save model
lgb_exp.bst.save_model(PATHS["lightgbm_model"])

# look at results manually
import shap
from s2aff.features import parse_ror_entry_into_single_string_lightgbm

ror_index = RORIndex()


with open(PATHS["lightgbm_gold_features"], "rb") as f:
    [
        (X_train, y_train, w_train, group_train, rors_train, qs_train),
        (X_val, y_val, w_val, group_val, rors_val, qs_val),
        (X_test, y_test, w_test, group_test, rors_test, qs_test),
    ] = pickle.load(f)


train_main_flag = np.array([i.endswith("[MAIN ONLY]") for i in qs_train])
train_main_address_flag = np.array([i.endswith("[MAIN+ADDRESS]") for i in qs_train])
train_augmented_flag = train_main_flag | train_main_address_flag

X_train = X_train[~train_augmented_flag]
y_train = y_train[~train_augmented_flag]
w_train = w_train[~train_augmented_flag]
group_train = group_train[~train_augmented_flag]
rors_train = rors_train[~train_augmented_flag]
qs_train = qs_train[~train_augmented_flag]

scores_train = lgb_exp.bst.predict(X_train)
scores_test = lgb_exp.bst.predict(X_test)
scores_val = lgb_exp.bst.predict(X_val)

lgb_exp.bst.params["objective"] = "rank_xendcg"
explainer = shap.Explainer(lgb_exp.bst)
shap_values_train = explainer.shap_values(X_train)
shap_values_val = explainer.shap_values(X_val)


def print_val(i, split="train"):
    if split == "train":
        flag = group_train == i
        q = qs_train[flag][0]
        rors = rors_train[flag]
        p = scores_train[flag]
        y_true = y_train[flag]
        x = X_train[flag]
    elif split == "val":
        flag = group_val == i
        q = qs_val[flag][0]
        rors = rors_val[flag]
        p = scores_val[flag]
        y_true = y_val[flag]
        x = X_val[flag]
    elif split == "test":
        flag = group_test == i
        q = qs_test[flag][0]
        rors = rors_test[flag]
        p = scores_test[flag]
        y_true = y_test[flag]
        x = X_test[flag]

    argsort = np.argsort(p)[::-1]  # largest to smallest
    rors = rors[argsort]
    p = p[argsort]
    y_true = y_true[argsort]
    x = x[argsort]

    print("Affiliation string:", q)
    print("----------")
    print("Correct ones:")
    for label, ror, prob in zip(y_true, rors, p):
        if label == 1:
            print(np.round(prob, 3), ror, parse_ror_entry_into_single_string_lightgbm(ror, ror_index))

    print("----------")
    print("Incorrect ones:")
    counter = 0
    for label, ror, prob in zip(y_true, rors, p):
        if label == 0:
            print(np.round(prob, 3), ror, parse_ror_entry_into_single_string_lightgbm(ror, ror_index))
            counter += 1
        if counter == 5:
            break


train_results = eval_lgb(y_train, X_train, lgb_exp.bst.predict(X_train), group_train)
val_results = eval_lgb(y_val, X_val, lgb_exp.bst.predict(X_val), group_val)
test_results = eval_lgb(y_test, X_test, lgb_exp.bst.predict(X_test), group_test)

wrong_ones_train = train_results["group_vec"][np.array(train_results["Prec@1_vec"]) == False]
wrong_ones_val = val_results["group_vec"][np.array(val_results["Prec@1_vec"]) == False]
wrong_ones_test = test_results["group_vec"][np.array(test_results["Prec@1_vec"]) == False]

for i in wrong_ones_train:
    print_val(i, "train")
    print("\n=====================================================\n")


# i = wrong_ones_val[-2]
# flag = group_val == i
# q = qs_val[flag][0]
# rors = rors_val[flag]
# p = scores_val[flag]
# y_true = y_val[flag]
# x = X_val[flag]
# shap_v = shap_values_val[flag]

i = wrong_ones_train[-5]
flag = group_train == i
q = qs_train[flag][0]
rors = rors_train[flag]
p = scores_train[flag]
y_true = y_train[flag]
x = X_train[flag]
shap_v = shap_values_train[flag]


argsort = np.argsort(p)[::-1]  # largest to smallest
rors = rors[argsort]
p = p[argsort]
y_true = y_true[argsort]
x = np.round(x[argsort], 3)
shap_v = np.round(shap_v[argsort], 3)

first_zero = np.where(y_true == 0)[0][0]
first_one = np.where(y_true == 1)[0][0]

print(q)
parse_ror_entry_into_single_string_lightgbm(rors[first_zero], ror_index)
list(zip(FEATURE_NAMES, x[first_zero], shap_v[first_zero]))
print()
parse_ror_entry_into_single_string_lightgbm(rors[first_one], ror_index)
list(zip(FEATURE_NAMES, x[first_one], shap_v[first_one]))



import re
import setuptools
from os import path

requirements_file = path.join(path.dirname(__file__), "requirements.in")
requirements = [r for r in open(requirements_file).read().split("\n") if not re.match(r"^\-", r)]

setuptools.setup(
    name="s2aff",
    version="0.21",
    url="https://github.com/allenai/S2AFF",
    packages=setuptools.find_packages(),
    install_requires=requirements,  # dependencies specified in requirements.in
    description="Semantic Scholar's Affiliation Extraction: Link Your Raw Affiliations to ROR IDs",
)



import pickle
from s2aff.model import NERPredictor
from s2aff.consts import PATHS

# load provided training data
# note: how this was made is not included in the release repo
# as it is messy and hard to reproduce
with open(PATHS["ner_training_data"], "rb") as f:
    df_training, df_validation, df_test = pickle.load(f)

ner_model = NERPredictor()

validation_metrics = ner_model.fit(df_training, df_validation)
print(validation_metrics)

# if you want to overwrite the provided model uncomment here
# ner_model.save_model(PATHS["ner_model"])



"""
How good is the first-stage ranker? Find out by running this script!

train + val + test:
MRR: 0.9081398880736107
Recall@25: 0.9751609935602575
Recall@100: 0.9829806807727691 <- reranking at this point (as of now)
Recall@250: 0.9875804967801288
Recall@500: 0.9894204231830727
Recall@1000: 0.9917203311867525
Recall@10000: 0.9931002759889604
Recall@100000: 0.9940202391904324
"""

import numpy as np
from s2aff.data import load_gold_affiliation_annotations
from s2aff.ror import RORIndex, index_min
from s2aff.model import NERPredictor, parse_ner_prediction


USE_CUDA = False

# load the gold, and subset to training data
df = load_gold_affiliation_annotations()
# df = df[df.split == "train"]

# load the ROR index and NERPredictor
ner_predictor = NERPredictor(use_cuda=USE_CUDA)  # defaults to models in the data directory

# get NER predictions for a bunch
ner_predictions = ner_predictor.predict(df.original_affiliation.values)


def run_stage_1_ranker(
    word_multiplier=0.3,
    ns={3},
    max_intersection_denominator=True,
    use_prob_weights=True,
    insert_early_candidates_ind=1,
    verbose=False,
):
    # load ror index
    ror_index = RORIndex(
        word_multiplier=word_multiplier,
        ns=ns,
        max_intersection_denominator=max_intersection_denominator,
        use_prob_weights=use_prob_weights,
        insert_early_candidates_ind=insert_early_candidates_ind,
    )

    # eval
    i = 0
    results = []
    for _, row in df.iloc[i:].iterrows():
        main, child, address, early_candidates = parse_ner_prediction(ner_predictions[i], ror_index)
        candidates, scores = ror_index.get_candidates_from_main_affiliation(main, address, early_candidates)
        gold_names = [ror_index.ror_dict[i]["name"] for i in row.labels]
        candidates_names = [ror_index.ror_dict[i]["name"] for i in candidates]
        rank_of_gold = index_min(candidates, row.labels)  # location of the gold in candidates
        result = {
            "original_affiliation": row.original_affiliation,
            "main": main,
            "child": child,
            "address": address,
            "early_candidates": early_candidates,
            "candidates": candidates,
            "candidate_names": candidates_names,
            "gold": row.labels,
            "gold_names": gold_names,
            "candidate_scores": scores,
            "gold_rank": rank_of_gold + 1,
            "reciprocal_rank": 1 / (rank_of_gold + 1),
        }

        results.append(result)
        mrr = np.mean([result["reciprocal_rank"] for result in results])
        if verbose:
            print(f"{i} | MRR: {mrr}")
        i += 1

    # aggregated scores
    mrr = np.mean([result["reciprocal_rank"] for result in results])

    metrics = {"mrr": mrr}
    if verbose:
        print(f"MRR: {mrr}")

    # recalls at various values
    for r_k in [25, 100, 250, 500, 1000, 10000, 100000]:
        r_k_recall = np.mean([result["gold_rank"] <= r_k for result in results])
        metrics[f"recall@{r_k}"] = r_k_recall
        if verbose:
            print(f"Recall@{r_k}: {r_k_recall}")

    return metrics, results, ror_index


metrics, results, ror_index = run_stage_1_ranker(verbose=True)

for k, v in metrics.items():
    print(f"{k}: {v}")

# find all the ones that don't have gold_rank == 1
results_bad = []
results_failed_recall = []
for result in results:
    if result["gold_rank"] != 1:
        results_bad.append(result)
    if result["gold_rank"] > 100:
        results_failed_recall.append(result)

for result in results_failed_recall:
    print("Original_affiliation:", result["original_affiliation"])
    print("Main/Child/Address", f"{result['main']} / {result['child']} / {result['address']}")
    print("Gold:", result["gold_names"])
    print("-------------------------------------------------")


# all_metrics = {}
# for word_multiplier in [0.3]:
#     for insert_early_candidates_ind in [0, 1, 2, 3, 4, 5, None]:
#         key = (word_multiplier, insert_early_candidates_ind)
#         if key not in all_metrics:
#             all_metrics[key] = run_stage_1_ranker(
#                 word_multiplier=word_multiplier, insert_early_candidates_ind=insert_early_candidates_ind, verbose=False
#             )
#         print(f"{key} | MRR: {all_metrics[key][0]['mrr']} | Recall@100: {all_metrics[key][0]['recall@100']}")

# for key in all_metrics.keys():
#     print(f"{key} | MRR: {all_metrics[key][0]['mrr']} | Recall@100: {all_metrics[key][0]['recall@100']}")



"""
This file contains the classes required by Semantic Scholar's
TIMO tooling.

You must provide a wrapper around your model, as well
as a definition of the objects it expects, and those it returns.
"""

from typing import List
from os.path import join, basename
import torch

from pydantic import BaseModel, BaseSettings, Field

from s2aff import S2AFF
from s2aff.ror import RORIndex
from s2aff.model import NERPredictor, PairwiseRORLightGBMReranker
from s2aff.consts import PATHS


class Instance(BaseModel):
    """
    Describes one Instance over which the model performs inference.

    The fields below are examples only; please replace them with
    appropriate fields for your model.

    To learn more about declaring pydantic model fields, please see:
    https://pydantic-docs.helpmanual.io/
    """

    raw_affiliation: str = Field(description="Raw affiliation string")



class Prediction(BaseModel):
    """
    Describes the outcome of inference for one Instance

    The fields below are examples only; please replace them with
    appropriate fields for your model.

    To learn more about declaring pydantic model fields, please see:
    https://pydantic-docs.helpmanual.io/
    """

    ror_id: str = Field(description="ROR id for the top result")
    display_name: str = Field(description="Standardized name for the top result")
    score: float = Field(description="Score from the LightGBM stage model for the top result")
    main: List[str] = Field(description="Main affiliation strings from NER step")
    child: List[str] = Field(description="Child affiliation strings from NER step")
    address: List[str] = Field(description="Address affiliation strings from NER step")

class PredictorConfig(BaseSettings):
    """
    Configuration required by the model to do its work.
    Uninitialized fields will be set via Environment variables.

    The fields below are examples only; please replace them with ones
    appropriate for your model. These serve as a record of the ENV
    vars the consuming application needs to set.
    """


class Predictor:
    """
    Interface on to your underlying model.

    This class is instantiated at application startup as a singleton.
    You should initialize your model inside of it, and implement
    prediction methods.

    If you specified an artifacts.tar.gz for your model, it will
    have been extracted to `artifacts_dir`, provided as a constructor
    arg below.
    """

    _config: PredictorConfig
    _artifacts_dir: str

    def __init__(self, config: PredictorConfig, artifacts_dir: str):
        self._config = config
        self._artifacts_dir = artifacts_dir
        self._load_model()

    def _load_model(self) -> None:
        """
        Perform whatever start-up operations are required to get your
        model ready for inference. This operation is performed only once
        during the application life-cycle.
        """
        ner_predictor = NERPredictor(
            model_path=join(self._artifacts_dir, basename(PATHS["ner_model"])), use_cuda=torch.cuda.is_available()
        )
        ror_index = RORIndex(
            ror_data_path=join(self._artifacts_dir, basename(PATHS["ror_data"])),
            country_info_path=join(self._artifacts_dir, basename(PATHS["country_info"])),
            works_counts_path=join(self._artifacts_dir, basename(PATHS["openalex_works_counts"])),
        )
        pairwise_model = PairwiseRORLightGBMReranker(
            ror_index,
            model_path=join(self._artifacts_dir, basename(PATHS["lightgbm_model"])),
            kenlm_model_path=join(self._artifacts_dir, basename(PATHS["kenlm_model"])),
        )
        self.s2aff = S2AFF(ner_predictor, ror_index, pairwise_model)

    @staticmethod
    def convert_raw_prediction_to_Prediction(prediction) -> Prediction:
        if len(prediction["stage2_candidates"]) == 0:
            prediction_instance = Prediction(ror_id="", score=0, main="", child="", address="")
        else:
            prediction_instance = Prediction(
                ror_id=prediction["stage2_candidates"][0],
                display_name=prediction["top_candidate_display_name"],
                score=prediction["stage2_scores"][0],
                main=prediction["main_from_ner"],
                child=prediction["child_from_ner"],
                address=prediction["address_from_ner"],
            )

        return prediction_instance

    def predict_batch(self, instances: List[Instance]) -> List[Prediction]:
        """
        Method called by the client application. One or more Instances will
        be provided, and the caller expects a corresponding Prediction for
        each one.

        If your model gets performance benefits from batching during inference,
        implement that here, explicitly.

        Otherwise, you can leave this method as-is and just implement
        `predict_one()` above. The default implementation here passes
        each Instance into `predict_one()`, one at a time.

        The size of the batches passed into this method is configurable
        via environment variable by the calling application.
        """
        predictions = self.s2aff.predict([i.raw_affiliation for i in instances])
        return [self.convert_raw_prediction_to_Prediction(i) for i in predictions]
